"""Base class for python path sources"""
# ruff: noqa: E501

from collections.abc import Mapping, Sequence
from pathlib import Path
from types import MappingProxyType
from typing import Optional, Union

from ... import _validate
from ..._load_strategy import _LoadStrategy
from ..._log import _log
from ..._marker_type import _MarkerType
from ..._path_resolution import _PathResolution
from ..._types import LoadStrategyLiterals, PathResolutionLiterals, RepoMarkerLiterals

__all__ = []


class NotPresent:
    """Sentinel class to represent a value that is not present."""


NOT_PRESENT = NotPresent()


class _Config:
    """Base configuration class for python path sources.


    :property repo_markers: Mapping of repository markers to their MarkerType.
    :property paths: Additional paths to include in :data:`sys.path`.
    :property load_strategy: The load strategy for handling multiple :data:`sys.path` sources.
    :property path_resolution_order: The order in which to resolve :data:`sys.path` sources.
    """

    __slots__ = ('_repo_markers', '_paths', '_load_strategy', '_path_resolution_order')

    def __init__(
        self,
        *,
        repo_markers: Optional[Mapping[str, Union[_MarkerType, RepoMarkerLiterals]]] = None,
        paths: Optional[Sequence[Union[Path, str]]] = None,
        load_strategy: Optional[Union[_LoadStrategy, LoadStrategyLiterals]] = None,
        path_resolution_order: Optional[Sequence[Union[_PathResolution, PathResolutionLiterals]]] = None,
        strict: bool = False,
    ) -> None:
        """
        :param Mapping[str, MarkerType | Literal['dir', 'file']] | None repo_markers: Markers to identify the repository root.
            Mapping of file or directory names to their MarkerType.

        :param Sequence[Path | str] | None paths: Additional paths to include in :data:`sys.path`.
            Sequence of paths relative to the repository root to be added to :data:`sys.path`.

        :param LoadStrategy | Literal['prepend', 'prepend_highest_priority', 'replace'] | None load_strategy: The
            strategy for loading :func:`sys.path` sources.

            It is expected to be one of `prepend`, `prepend_highest_priority`, or `replace` (as defined
            in :class:`LoadStrategy`). It can use either the enum value or its string representation.

        :param Sequence[PathResolution | Literal['manual', 'autopypath', 'pyproject']]] | None path_resolution_order: The order in which to
            resolve :func:`sys.path` sources.

            It is expected to be a sequence containing any of the following values:
            `manual`, `autopypath`, `pyproject` as defined in :class:`PathResolution`.
            It can use either the enum values or their string representations.
        :param bool strict: (default: ``False``) Indicates whether strict mode is enabled for error handling.
        """
        self._repo_markers: Union[MappingProxyType[str, _MarkerType], None] = _validate.repo_markers(repo_markers)
        self._paths: Union[tuple[Path, ...], None] = _validate.paths(paths)
        self._load_strategy: Union[_LoadStrategy, None] = _validate.load_strategy(load_strategy)
        self._path_resolution_order: Union[tuple[_PathResolution, ...], None] = _validate.path_resolution_order(
            path_resolution_order
        )

        cls = self.__class__.__name__
        _log.debug(
            f'{cls} initialized with repo_markers={self._repo_markers}, '
            f'paths={self.paths}, load_strategy={self.load_strategy}, '
            f'path_resolution_order={self.path_resolution_order}'
        )

    @property
    def repo_markers(self) -> Union[MappingProxyType[str, _MarkerType], None]:
        """Mapping of repository markers to their MarkerType.

        :return MappingProxyType[str, MarkerType] | None: A mapping where keys are filenames or directory names
            that indicate the repository root, and values are of type `MarkerType`.
            ``None`` if no custom repo markers are set.
        """
        return self._repo_markers

    @property
    def load_strategy(self) -> Union[_LoadStrategy, None]:
        """The load strategy for handling multiple :func:`sys.path` sources.

        :return LoadStrategy | None: The strategy used when handling multiple sys.path sources.
            ``None`` if no custom load strategy is set.
        """
        return self._load_strategy

    @property
    def path_resolution_order(self) -> Union[tuple[_PathResolution, ...], None]:
        """The order in which to resolve :func:`sys.path` sources.

        :return tuple[PathResolution, ...] | None: A tuple defining the order of resolution for sys.path sources.
            ``None`` if no custom resolution order is set.
        """
        return self._path_resolution_order

    @property
    def paths(self) -> Union[tuple[Path, ...], None]:
        """Additional paths to include in :data:`sys.path`.

        :return tuple[Path, ...] | None: A tuple of additional paths relative to the repository root
            to be added to :data:`sys.path`. ``None`` if no additional paths are set.
        """
        return self._paths

    def __repr__(self) -> str:
        """String representation of the Config object.

        :return str: A string representation of the Config instance.
        """
        cls = self.__class__.__name__

        paths_list: list[str] = []
        if self.paths is not None:
            for path in self.paths:
                paths_list.append(f'{str(path)!r}')
            path_repr = '[' + ', '.join(paths_list) + ']'
        else:
            path_repr = 'None'

        markers_list: list[str] = []
        if self.repo_markers is not None:
            for key, value in self.repo_markers.items():
                marker_str = f'_MarkerType.{value.name}'
                markers_list.append(f'{key!r}: {marker_str}')
            markers_repr = '{' + ', '.join(markers_list) + '}'
        else:
            markers_repr = 'None'

        path_resolution_list: list[str] = []
        if self.path_resolution_order is not None:
            for order in self.path_resolution_order:
                path_resolution_list.append(f'_PathResolution.{order.name}')
            path_resolution_repr = '[' + ', '.join(path_resolution_list) + ']'
        else:
            path_resolution_repr = 'None'

        load_strategy_repr = f'_LoadStrategy.{self.load_strategy.name}' if self.load_strategy is not None else 'None'

        return (
            f'{cls}(repo_markers={markers_repr}, '
            f'paths={path_repr}, load_strategy={load_strategy_repr}, '
            f'path_resolution_order={path_resolution_repr})'
        )

    def __eq__(self, other: object) -> bool:
        """Equality comparison for Config objects.

        :param object other: Another object to compare with.
        :return bool: True if both Config instances are equal, False otherwise.
        """
        if not isinstance(other, _Config):
            return NotImplemented
        return (
            self.repo_markers == other.repo_markers
            and self.paths == other.paths
            and self.load_strategy == other.load_strategy
            and self.path_resolution_order == other.path_resolution_order
        )

    def __hash__(self) -> int:
        """Hash for Config objects."""
        repo_markers: MappingProxyType[str, _MarkerType] = self.repo_markers or MappingProxyType({})
        return hash(
            (frozenset(sorted(repo_markers.items())), self.paths, self.load_strategy, self.path_resolution_order)
        )

    def replace(
        self,
        *,
        repo_markers: Union[Mapping[str, Union[_MarkerType, RepoMarkerLiterals]], None, NotPresent] = NOT_PRESENT,
        paths: Union[Sequence[Union[Path, str]], None, NotPresent] = NOT_PRESENT,
        load_strategy: Union[_LoadStrategy, LoadStrategyLiterals, None, NotPresent] = NOT_PRESENT,
        path_resolution_order: Union[
            Sequence[Union[_PathResolution, PathResolutionLiterals]], None, NotPresent
        ] = NOT_PRESENT,
    ) -> '_Config':
        """Creates a copy of the current Config instance with specified attributes replaced.

        If an attribute is not provided, the value from the current instance is used.

        The default value for each parameter is a sentinel `NotPresent` instance,
        which indicates that the current value should be retained.

        :param Mapping[str, MarkerType] | None | NotPresent repo_markers: New repo_markers value.
        :param Sequence[Path | str] | None | NotPresent paths: New paths value.
        :param LoadStrategy | str | None | NotPresent load_strategy: New load_strategy value.
        :param Sequence[PathResolution | str] | None | NotPresent path_resolution_order: New path_resolution_order
            value.

        :return Config: A new Config instance with selected attributes replaced.
        """

        new_repo_markers = self.repo_markers if isinstance(repo_markers, NotPresent) else repo_markers
        new_paths = self.paths if isinstance(paths, NotPresent) else paths
        new_load_strategy = self.load_strategy if isinstance(load_strategy, NotPresent) else load_strategy
        new_path_resolution_order = (
            self.path_resolution_order if isinstance(path_resolution_order, NotPresent) else path_resolution_order
        )
        return _Config(
            repo_markers=new_repo_markers,
            paths=new_paths,
            load_strategy=new_load_strategy,
            path_resolution_order=new_path_resolution_order,
        )



"""Module providing autopypath functionality.

It allows automatic management of Python paths for project scripts so that
they can be run directly without manual path adjustments or reliance on IDE features.

It is **NOT** meant to replace normal package management or virtual environments,
but to facilitate running scripts directly during development and testing when
the build may be broken due to import errors in unrelated code that prevent the normal test
runner from executing at all (:mod:`pytest`, for example, will refuse to run tests if import errors occur
even in unrelated modules).

It detects the project root by looking for specific marker files or directories
(such as `pyproject.toml`, `.git`, or `.hg`) and adjusts `sys.path` to include
relevant subdirectories based on configurations found in `autopypath.toml` or
`pyproject.toml` files.

It recursively searches parent directories looking for marker files and directories
to identify the project root. Once found, it reads configuration settings to determine
which paths to add to `sys.path`, how to load them, and in what order.

Configuration Sources and Precedence
====================================

The module supports multiple configuration sources for determining which paths
to add to `sys.path`:

**direct manual configuration**
-------------------------------

Users can specify repository markers, manual paths, load strategies, and
path resolution orders directly via the :func:`autopypath.custom.configure_pypath` function.

This method has the highest precedence and overrides any file-based configurations
that are set. It allows users to fully customize the behavior of autopypath
in a single script for their specific project structure and requirements without
relying on configuration files.

Options that are NOT specified in the direct manual configuration will fall back
to the next available configuration source in the precedence order.

**autopypath.toml**
-------------------

If present in a directory found during the root search, this file takes highest
precedence for the file-based configurations but not over direct manual configuration.

It can specify custom repository markers, paths, load strategies, and resolution orders.
Again, any options not specified here will fall back to the next available configuration source.

This is not meant to replace `pyproject.toml` but to provide a dedicated
configuration file for autopypath settings that may need to be set before
a `pyproject.toml` can even be located or parsed.

A typical use case is from a script inside a project subdirectory such as ``tests/``
but the project root does not contain one of the default marker files (e.g.,
``pyproject.toml``, ``.git/``).

In this case, an ``autopypath.toml`` file can be placed in the subdirectory
to specify the necessary configuration to find the project root and paths.

This allows for early configuration of the search for the project root and paths.

While it can coexist with `pyproject.toml`, it is primarily intended for use
cases where early configuration is necessary and not as a replacement for `pyproject.toml`.

If it finds a `autopypath.toml` file while searching for the project root, it uses that
configuration to update the marker files, paths, load strategy, and path resolution order
but does not halt the search for the project root unless that directory also satisfies
the newly updated markers - it continues searching upwards until it finds a recognized marker
file or directory. It is not limited to using only the directory containing the
`autopypath.toml` file as the project root.

This is a special behavior unique to `autopypath.toml` files where it can influence
the search for the project root itself during the upward search. Because it can change
the markers used to identify the project root, it may cause the search to continue
upwards beyond the directory containing the `autopypath.toml` file even if that directory
would otherwise be considered the project root based on the original markers.

For example, if the original markers were `pyproject.toml` and `.git`, but the `autopypath.toml` file changes
the markers to only `.hg`, it will continue searching until it finds a `.hg` directory
to use as the project root but will ignore any `pyproject.toml` or `.git`
directories found along the way.

e.g., if `autopypath.toml` specifies the `.git` directory as a marker, but
the `.git` directory is located in a parent directory, it will continue
searching upwards until it finds that `.git` directory and uses that as the project root.

Scope Limitation
~~~~~~~~~~~~~~~~

Only the first `autopypath.toml` file found during the upward search is used;
any others in parent directories are ignored. This scope-limited approach prevents
conflicts from multiple configuration files in nested projects.

This allows for flexible configuration at different levels of a project hierarchy and
supports monorepos or multi-package repositories.

**pyproject.toml**
------------------

It looks for a `pyproject.toml` file in the project root directory. If it contains
a `[tool.autopypath]` section, it uses that configuration unless overridden
by direct manual configuration or `autopypath.toml`.

**default configuration**
-------------------------

If no other configurations are found, it falls back to a default configuration
that sets common source directories like `src/` and `lib/` to include in `sys.path` and
uses a standard load strategy and path resolution order and set of common repository markers
(e.g., `pyproject.toml`, `.git`, etc.) to find the project root.

The default markers, paths, load strategy, and resolution order can be found in the
:mod:`autopypath.defaults` module.

For easy reference, here are the defaults:

**Default Repository Markers**
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- `pyproject.toml` (file)
- `autopypath.toml` (file)
    Note that there is special behavior for this marker if found. If it changes
    the repo_markers settings, the repo root is re-evaluated using the new settings
    and that may result in a different repository root being identified.
- `.git/` (directory, Git version control)
- `.hg/` (directory, Mercurial version control)
- `.svn/` (directory, Subversion version control)
- `.bzr/` (directory, Bazaar version control)
- `.cvs/` (directory, CVS version control)
- `_darcs/` (directory,  Darcs version control)
- `.fossil/` (directory, Fossil version control)

**Default Paths to Add**
~~~~~~~~~~~~~~~~~~~~~~~~
- `src/`
- `tests/`

**Default Load Strategy**
~~~~~~~~~~~~~~~~~~~~~~~~~

- `prepend` (loads all specified paths, merging them and prepending them to `sys.path`)

**Default Path Resolution Order**
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

1. `manual` (uses any manually specified paths first)
2. `autopypath` (appends paths from `autopypath.toml` file after manual paths)
3. `pyproject` (appends paths from `pyproject.toml` file after `autopypath.toml` paths)

Behavior When No Valid Configuration is Found
=============================================

If no valid project root or paths are found, it does not modify `sys.path` at all
but instead logs warnings to inform the user. It only makes adjustments when it can
confidently determine the correct paths to add and has been explicitly invoked
directly from a script being run as `__main__`. It will not add paths to non-existent
directories.

If there is any doubt about whether to adjust the path, the module opts to do nothing.

This is to avoid interfering with normal development workflows and to 'fail to safe'
by preventing accidentally breaking the environment. A 'if in doubt, do nothing' approach.

.. note::

   The automatic configuration only occurs when the module is imported
   at the top level of a script that is run directly (i.e., `__name__ == "__main__"`).

   It cannot be used from inside functions or class definitions. If the module is imported
   in any other context, it will log debug information but will not adjust `sys.path`.

   It should be imported before any other imports that depend on the adjusted Python path.

   The custom configuration function also only works when called directly
   from the top level of a script being executed as `__main__` - not from within
   functions or classes.

   Again, the function should be called before any other imports that depend on the adjusted Python path.

Modes of Operation
==================

**Fully Automatic Mode**
------------------------

When imported from a script, it automatically adjusts `sys.path` based on
`autopypath.toml` file or a `pyproject.toml` file.

This ensures that modules can be imported correctly when running scripts directly.

.. code-block:: python

    import autopypath
    # sys.path is now adjusted automatically

**Fully Automatic Debug Mode**
------------------------------

By importing the `autopypath.debug` submodule,
detailed debug logging is enabled to trace how the project root is determined,
which paths are added to `sys.path`, and any issues encountered along the way.

This is useful for troubleshooting and understanding the internal workings of autopypath.

.. code-block:: python

    import autopypath.debug
    # sys.path is now adjusted automatically with debug logging enabled

**Custom Configured Mode**
--------------------------

Users can call the :func:`autopypath.custom.configure_pypath`
function to specify custom repository markers, manual paths, and loading strategies.
This provides flexibility for different project structures and requirements.

The path is **NOT** adjusted automatically on import in this mode; the user must call the
:func:`configure_pypath` function explicitly. It can only be called from the top level
of a script being executed as `__main__`, not from within functions or classes.
It will not work if called from other contexts. The prevents it from accidentally
modifying `sys.path` except when explicitly intended (such as when running a script directly
rather than being imported as a module such as by pytest or other test runners).

It should be called before any other imports that depend on the adjusted Python path.

.. code-block:: python

    from autopypath.custom import configure_pypath

    configure_pypath(
        repo_markers={'setup.py': 'file', '.git': 'dir'},
        manual_paths=['src', 'lib'],
        load_strategy='prepend',
    )
    # sys.path is now adjusted based on custom configuration

**Custom Configured Debug Mode**
--------------------------------

By importing the `autopypath.custom.debug` submodule,
users can enable detailed debug logging while using custom configuration options.

The path is **NOT** adjusted automatically on import; the user must call the
:func:`configure_pypath` function explicitly.

.. code-block:: python

    from autopypath.custom.debug import configure_pypath

    configure_pypath(
        repo_markers={'setup.py': 'file', '.git': 'dir},
        manual_paths=['src', 'lib'],
        load_strategy='prepend',
    )
    # sys.path is now adjusted based on custom configuration

"""

from pathlib import Path
from typing import Optional

from ._config_py_path import _ConfigPyPath
from ._context import _context_frameinfo
from ._exceptions import AutopypathError
from ._log import _log

__all__ = [
    'AutopypathError',
]


# Only run if directly imported by a script being executed as __main__
# If there is any doubt, does not run automatically

_context_file: Optional[Path] = None
"""This is the file path of the script that imported this module, if available."""
_context_name: Optional[str] = None
"""This is the __name__ of the script that imported this module, if available."""
_path_adjusted: bool = False
"""Indicates whether autopypath adjusted sys.path automatically upon import."""

_context_info: Optional[tuple[Path, str]] = _context_frameinfo()
if _context_info is not None:
    _context_file, _context_name = _context_info
    if _context_name != '__main__':
        _log.debug(
            'autopypath imported from non-__main__ context (%s); no sys.path changes will be applied.',
            _context_name,
        )
        _path_adjusted = False
    else:
        _ConfigPyPath(context_file=_context_file)
        _log.debug('sys.path adjusted automatically for %s', _context_file)
        _path_adjusted = True
else:
    _log.error('could not determine context file; no sys.path changes will be applied.')
    raise AutopypathError('could not determine context file for autopypath; sys.path not adjusted.')



"""TestSpec framework - TestSetGet class"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from types import TracebackType
from typing import Any, Callable, NoReturn, Optional, Union

from .assertions import Assert, validate_assertion
from .base import TestSpec
from .constants import NO_EXPECTED_VALUE, NO_OBJ_ASSIGNED
from .helpers import _process_exception


@dataclass
class TestSetGet(TestSpec):
    """A class for testing setting and getting attributes.

    This class allows for testing the setting and getting of attributes on an object,
    including validation of the set value, expected exceptions, and custom validation functions.

    :param name: Identifying name for the test.
    :type name: str
    :param obj: The object whose attribute is to be tested. If not provided, the special sentinel value
                NO_OBJ_ASSIGNED is used. The object must not be None and must be an instance of object.
                It must be provided to run the test.
                Defaults to NO_OBJ_ASSIGNED.
    :type obj: Optional[object], optional
    :param attribute: The name of the attribute to be tested by setting.
    :type attribute: str
    :param value: Value to set the attribute to.
    :type value: Any
    :param expected: Expected value of attribute after setting the attribute. If a get_exception or
                     set_exception is set, the expected value is ignored. If there is no expected exception
                     and no expected value, use the special sentinel value NO_EXPECTED_VALUE to skip the
                     get step validation for the set value.
                     Defaults to NO_EXPECTED_VALUE.
    :type expected: Optional[Any], optional
    :param get_attribute: The name of the attribute to be retrieved for 'expected' validation.
                          If None, use the same as `attribute`.
                          Defaults to None.
    :type get_attribute: Optional[str], optional
    :param assertion: The assertion operator to use when comparing the expected and found values.
                      Defaults to Assert.EQUAL.
    :type assertion: Assert, optional
    :param exception: Expected exception type (if any) to be raised by setting the attribute.
                      Defaults to None.
    :type exception: Optional[type[Exception]], optional
    :param exception_tag: Expected tag (if any) to be found in the exception message.
                          Defaults to None.
    :type exception_tag: Optional[str | Enum], optional
    :param validate: Function to validate obj after setting the attribute.
                     It should return True if the object state is valid.
                     This is distinguished from the expected value check in that it can perform more complex validation
                     of the entire object state rather than just checking the value of a single attribute.
                     It is passed two arguments, the TestSetGet instance and the object being validated.
                     Defaults to None.
    :type validate: Optional[Callable[[TestSetGet, Any], bool]], optional
    :param on_fail: Function to call on test failure to raise an exception.
                    Defaults to pytest.fail.
    :type on_fail: Callable[[str], NoReturn], optional
    :param extra: Extra fields for use by test frameworks. It is not used by the TestSetGet class itself.
                  Defaults to None.
    :type extra: Any, optional
    """

    __test__ = False  # Prevent pytest from trying to collect this class as a test case

    name: str
    """Identifying name for the test."""
    attribute: str
    """The name of the attribute to be tested by setting."""
    value: Any
    """Value to set the attribute to."""
    obj: Optional[object] = NO_OBJ_ASSIGNED
    """The object whose attribute is to be tested. It cannot be None, and must be an instance of object.
    If not provided during construction, the special sentinel value NO_OBJ_ASSIGNED is used. This must be
    replaced with a valid object before running the test."""
    assertion: Assert = Assert.EQUAL
    """The assertion operator to use when comparing the expected and found values. (default is Assert.EQUAL)"""
    expected: Optional[Any] = NO_EXPECTED_VALUE
    """Expected value of attribute after setting the attribute. If a get_exception or
    set_exception is set, the expected value is ignored. If there is no expected exception
    and no expected value, use the special sentinel value NO_EXPECTED_VALUE to skip the
    get step validation for the set value. Omitting this field is equivalent to setting it to NO_EXPECTED_VALUE."""
    get_attribute: Optional[str] = None
    """The name of the attribute to be retrieved for 'expected' validation. If None, use the same as `attribute`."""
    set_exception: Optional[type[Exception]] = None
    """Expected exception type (if any) to be raised by setting the attribute."""
    set_exception_tag: Optional[Union[str, Enum]] = None
    """Expected tag (if any) to be found in an exception message raised by setting the attribute."""
    get_exception: Optional[type[Exception]] = None
    """Expected exception type (if any) to be raised by getting the attribute."""
    get_exception_tag: Optional[Union[str, Enum]] = None
    """Expected tag (if any) to be found in an exception message raised by getting the attribute."""
    validate: Optional[Callable[[TestSetGet, Any], None | NoReturn]] = None
    """Function to validate obj state after setting the attribute. It should raise an exception
    if the object state is unexpected.

    This is distinguished from the expected value check in that it can perform more complex validation
    of the entire object state rather than just checking the value of a single attribute.

    It is passed two arguments, the TestSetGet instance and the object being validated.

    The validation function should call the `on_fail` method to raise an exception if the object is not
    in a valid state. None should be returned if the object is valid.
    """
    on_fail: Callable[[str], NoReturn] | None = None
    """Function to call on test failure. The function should raise an exception (default is _fail method)."""

    extra: Any = None
    """Extra data for use by test frameworks. It is not used by the TestSetGet class itself. Default is None."""

    _creation_traceback: Optional[TracebackType] = None
    """The traceback at the point where the TestAction was created."""

    def __post_init__(self) -> None:
        """Post-initialization validation checks."""
        if not isinstance(self.name, str):
            raise TypeError('name must be a str')
        if not isinstance(self.attribute, str):
            raise TypeError('attribute must be a str')
        if self.obj is None:
            raise TypeError('obj cannot be None')
        if not isinstance(self.obj, object):
            raise TypeError('obj must be an object')
        if self.attribute == '':
            raise ValueError('attribute cannot be an empty string')
        if self.validate is not None and not callable(self.validate):
            raise TypeError('validate must be callable if provided')
        if self.set_exception is not None and not issubclass(self.set_exception, Exception):
            raise TypeError('set_exception must be an Exception type if provided')
        if self.set_exception_tag is not None and not isinstance(self.set_exception_tag, (str, Enum)):
            raise TypeError('set_exception_tag must be a str or Enum if provided')
        if not callable(self.on_fail):
            raise TypeError('on_fail must be callable')
        super().__post_init__()

    def run(self) -> None:
        """Execute the attribute set/get test."""
        # disabled because we are using the __setattr__ and __getattribute__ dunder methods directly
        # for testing purposes because there is no other way to do get/set testing for attributes generically.
        # pylint: disable=unnecessary-dunder-call

        # hide traceback for this function in pytest output. Disabled for pylint because it is a pytest
        # feature that is just not understood by pylint.
        __tracebackhide__ = True  # pylint: disable=unused-variable

        test_description: str = f'{self.name}'

        # Errors found during the test
        errors: list[str] = []

        obj = self.obj
        if obj is NO_OBJ_ASSIGNED:
            if self.on_fail:
                self.on_fail(f'{self.name}: obj for test is not assigned')
            else:
                self._fail(f'{self.name}: obj for test is not assigned')
            raise RuntimeError('unreachable code after on_fail call')  # pylint: disable=raise-missing-from

        expected = self.expected
        validate = self.validate

        # Set the attribute and check for exceptions as appropriate
        try:
            if not hasattr(obj, self.attribute):
                errors.append(f'obj has no attribute {self.attribute}')
            else:
                obj.__setattr__(self.attribute, self.value)
                if self.set_exception is not None:
                    errors.append('set operation returned instead of raising an expected exception')

        except BaseException as err:  # pylint: disable=broad-exception-caught
            new_errors = _process_exception(
                err=err,
                exception=self.set_exception,
                exception_tag=self.set_exception_tag,
                label=f'setting attribute {self.attribute}',
            )
            errors.extend(new_errors)

        # bail now if there was an error during the set operation
        if errors:
            if self.on_fail:
                self.on_fail(test_description + ': ' + '\n'.join(errors))
            else:
                self._fail(test_description + ': ' + '\n'.join(errors))
            raise RuntimeError('unreachable code after on_fail call')  # pylint: disable=raise-missing-from

        # If there is no get_exception or expected value or validate function, we can't do any validation
        # on the set value, so just return now. This amounts to a minimal test of just setting the attribute
        # it causing an exception.
        if self.get_exception is None and expected is NO_EXPECTED_VALUE and validate is None:
            return

        # Perform post-set validations if requested. This is skipped if there was an exception
        # during the set operation.
        attribute_to_get = self.attribute if self.get_attribute is None else self.get_attribute
        try:
            if not hasattr(obj, attribute_to_get):
                errors.append(f'obj has no attribute {attribute_to_get}')
            else:
                if validate is not None:
                    validate(self, obj)  # Exception should be raised by validate if unexpected obj state

                if expected is NO_EXPECTED_VALUE:
                    return

                found: Any = obj.__getattribute__(attribute_to_get)
                assertion_error = validate_assertion(self.assertion, expected, found)
                if assertion_error:
                    errors.append(f"{assertion_error} for attribute '{attribute_to_get}'")

        except BaseException as err:  # pylint: disable=broad-exception-caught
            new_errors = _process_exception(
                err=err,
                exception=self.get_exception,
                exception_tag=self.get_exception_tag,
                label=f'getting attribute {attribute_to_get}',
            )
            errors.extend(new_errors)

        # Report any errors found during the validate portion of the test
        if errors:
            if self.on_fail:
                self.on_fail(test_description + ': ' + '\n'.join(errors))
            else:
                self._fail(test_description + ': ' + '\n'.join(errors))
            raise RuntimeError('unreachable code after on_fail call')  # pylint: disable=raise-missing-from



"""Base TestSpec class"""
import inspect
from abc import ABC, abstractmethod
from types import TracebackType
from typing import Union


class TestSpec(ABC):
    """Base class for test specifications."""
    # Prevent pytest from trying to collect this class as a test case
    __test__ = False

    _creation_traceback: Union[TracebackType, None] = None
    """The traceback at the point where the TestSpec was created."""

    @abstractmethod
    def run(self) -> None:
        """Run the test based on the provided TestSpec entry.

        This function is intended to be overridden by subclasses to implement
        specific test execution logic.
        """
        raise NotImplementedError("Subclasses must implement the run method.")

    def __post_init__(self) -> None:
        """Capture the traceback at the point of creation."""
        # Don't capture the traceback if this is a subclass, as it will be captured by the parent
        current_frame = inspect.currentframe()
        if current_frame is None:
            # This should not be able to happen in CPython, but it's better to be safe.
            raise RuntimeError("Could not get current frame.") from None

        try:
            while current_frame is not None:
                module = inspect.getmodule(current_frame)
                if module is None or module.__name__.startswith('testspec'):
                    current_frame = current_frame.f_back
                    continue

                # We've found the first frame outside the testspec package.
                self._creation_traceback = TracebackType(
                    None, current_frame, current_frame.f_lasti, current_frame.f_lineno
                )
                return
        finally:
            # As per the inspect module documentation, we need to delete the frame reference
            # when we are done with it to avoid reference cycles.
            del current_frame

        raise RuntimeError("Could not find creation traceback.") from None

    def _fail(self, message: str) -> None:
        """Raise a test failure, attaching the creation traceback if available."""
        __tracebackhide__ = True  # pylint: disable=unused-variable
        exc = AssertionError(message)
        if self._creation_traceback:
            raise exc.with_traceback(self._creation_traceback) from None
        # If we failed to get a traceback, raise the assertion as-is.
        raise exc



"""TestSpec testing framework."""

from .assertions import Assert
from .base import TestSpec
from .constants import NO_EXPECTED_VALUE, NO_OBJ_ASSIGNED
from .context import Context
from .helpers import no_assigned_action
from .idspec import idspec
from .pytest_action import PytestAction
from .test_action import TestAction
from .test_get import TestGet
from .test_set import TestSet
from .test_setget import TestSetGet

__test__ = False  # Prevent pytest from trying to collect this package as a test case

__all__ = [
    'NO_EXPECTED_VALUE',
    'NO_OBJ_ASSIGNED',
    'idspec',
    'no_assigned_action',
    'Assert',
    'Context',
    'PytestAction',
    'TestAction',
    'TestGet',
    'TestSet',
    'TestSetGet',
    'TestSpec',
]



"""Special types used in AutoPyPath.

This module provides compatibility imports for typing features
across different Python versions that may not have them natively.

The types provided here include:
- :class:`Literal`
- :class:`Final`
- :class:`Never`
- :class:`TypeAlias`
- :class:`TypeGuard`

"""
# ruff: noqa: F401

__all__ = []

# Compatibility imports for typing features
# This is a 'feature detection' pattern to support multiple Python versions
# without having to know exactly which versions support which features.
try:
    from typing import Literal  # type: ignore[attr-defined]
except ImportError:  # pragma: no cover  # Will only occur on very old Python versions
    from typing_extensions import Literal

try:
    from typing import Final  # type: ignore[attr-defined]
except ImportError:  # pragma: no cover  # Will only occur on very old Python versions
    from typing_extensions import Final

try:
    from typing import TypeAlias  # type: ignore[attr-defined]
except ImportError:  # pragma: no cover  # Will only occur on very old Python versions
    from typing_extensions import TypeAlias

try:
    from typing import TypeGuard  # type: ignore[attr-defined]
except ImportError:  # pragma: no cover  # Will only occur on very old Python versions
    from typing_extensions import TypeGuard

try:
    from typing import Never  # type: ignore[attr-defined]
except ImportError:  # pragma: no cover  # Will only occur on very old Python versions
    from typing_extensions import Never



"""Zero-dependency, cross-platform bootstrap script to set up a Python
development environment.

This is the development bootstrap script for the 'autopypath' project.

The bootstrap script is NOT intended to be run with Python versions earlier than 3.10
and will exit with an error if attempted.

It has been tested with Python 3.10 through 3.14 as of January 2025.

This script is part of the 'python-env-bootstrap' project available at
https://github.com/JerilynFranz/python-env-bootstrap

License
-------

Licensed under the Apache License, Version 2.0 (SPDX-License-Identifier: Apache-2.0)
https://www.apache.org/licenses/LICENSE-2.0.txt
Copyright [2025] Jerilyn Franz

Based on 'python-env-bootstrap' (https://github.com/JerilynFranz/python-env-bootstrap).

Description
-----------

It is designed to be run after cloning a git or Mercurial repository, to create
a local virtual environment (.venv for example), and install necessary development tools.

It relies only on the Python standard library and network access to PyPI and
does not require any pre-installed packages or change your system Python installation.

This script installs the following tools by default:
- uv (for managing Python packages and dependencies)

That is the minimum set of tools required to start development for this project.

The minimum supported Python version, virtual environment directory name,
and list of tools to install can be customized by modifying the
corresponding constants in this script. The script can also be extended to
perform additional setup steps as needed via the run_post_install_steps() function
such as installing additional packages from requirements.txt files
or configuring settings.

Settable Options
----------------

- VENV_DIR: The name of the virtual environment directory to create.
- BOOTSTRAP_MODULES: A list of InstallSpec instances specifying the packages
  to install into the virtual environment during bootstrap.
- Post-install steps: You can customize the `run_post_install_steps()`
  function to perform additional setup tasks after installing the core tools.
- Output control: You can set `DEFAULT_DEBUG` and `DEFAULT_QUIET` constants
  to control whether debug output or quiet mode is enabled by default.
- Command-line options: You can use '--debug'/'--no-debug' and '--quiet'/'--verbose'
  to control output verbosity when running the script.
- Automatic confirmation: You can use '--yes'/'-y' to skip confirmation prompts.
- You can configure the supported Python versions by modifying
  the version check ("if sys.version_info") at the start of the script.

Usage
-----

python bootstrap.py [-h] [--yes] [--debug | --no-debug] [-q | -v]

CLI Help
--------
  -h, --help     show this help message and exit
  --yes, -y      Automatically confirm and proceed without prompting.
  --debug        Enable debug output.
  --no-debug     Disable debug output.
  -q, --quiet    Suppress non-error output.
  -v, --verbose  Enable verbose output (default).
"""

__version__ = '1.3.0'

# pylint: disable=wrong-import-position,C0103,too-many-lines

import sys

# Check for minimum supported Python version before importing anything else
# this ensures that users get a clear error message if they try to run
# the script with an unsupported Python version.
#
# The minimum version of Python this bootstrap script can support is 3.10+
# This can be changed as needed for your project.
if sys.version_info < (3, 10):
    major, minor = sys.version_info.major, sys.version_info.minor
    print(f'Error: Python 3.10 or later is required to run this script. You are using Python {major}.{minor}.')
    sys.exit(2)

import argparse
import os
import shutil
import stat
import subprocess
from collections.abc import Callable
from functools import cache
from pathlib import Path
from typing import Any, NamedTuple
from venv import create as create_venv

DEFAULT_DEBUG: bool = False
"""Enable debug output only if --debug is specified.

To enable debug output by default, set this to True
and then --no-debug can be used to disable it.
"""

DEFAULT_QUIET: bool = False
"""Suppress non-error output only if --quiet is specified.

To enable quiet output by default, set this to True
and then --verbose can be used to disable it.
"""

# Whether to remove the bootstrapvirtual environment directory on script exit.
# This is useful for cleaning up the bootstrap venv after installation
# if desired. Set to False to keep the venv for inspection or reuse.
#
# Removing it on exit is not the default behavior to avoid accidental loss
# of the created environment. But it can be enabled as needed.
# Default is False because most users will want to keep the venv.
REMOVE_BOOTSTRAP_VENV_ON_EXIT: bool = False
"""Whether to remove the virtual environment directory on script exit."""

VENV_DIR: str = '.venv'
"""The name of the virtual environment directory to create in the repository root for the bootstrap."""

# The name of the final virtual environment directory when activated.
#
# This can be changed as needed for your project. If you are using 'tox'
# or another tool such as 'poetry' that creates its own venv,
# you may want to change this to 'venv' or another appropriate name depending on
# your project's virtual environment management.
ACTIVATED_VENV_DIR: str = '.venv'
"""The name of the virtual environment directory where the project's virtual environment is kept."""


class InstallSpec(NamedTuple):
    """Specification for modules required to be installed.

    :param str name: The name of the module to install.
    :param str version: An optional version specifier (e.g., ">=1.0.0").
    :param str extras: An optional extras specifier (e.g., "[dev]").
    """

    name: str
    version: str = ''
    extras: str = ''

    def __str__(self) -> str:
        return f'{self.name}{self.extras}{self.version or " (latest)"}'


# --- Modules to install during bootstrap ---

BOOTSTRAP_MODULES: list[InstallSpec] = [
    InstallSpec(name='uv', version='>=0.9.18'),
]

# --- Tool usage instructions template ---

TOOL_USAGE_INSTRUCTIONS = """
You use 'tox' to run tasks that set up and manage the development environment,
run tests, linters, and build documentation:

Examples:

  tox run -e lint     # Run linters on the codebase
  tox run -e docs     # Build documentation
  tox run -e py310    # Run the test suite using Python 3.10
  tox run -e py314    # Run the test suite using Python 3.14
  tox devenv -e dev   # Start an interactive dev environment with Python 3.12

The list of available 'tox' environments can be found by running:

  tox list

If you are not familiar with using 'tox' see https://tox.wiki/en/latest/

You use 'uv' to manage Python packages within the virtual environment and to
update pyproject dependencies:

Examples:

  # Add a new package to the 'dev' dependency group
  uv add --dev --group=dev 'package_name>=1.2.3'

  # Add a new package to the default dependency group
  uv add 'package_name>=1.2.3'

  # Add a package to specified extras
  uv add 'package_name[extra1,extra2]'

  # install a package from PyPI to the virtual environment
  uv pip install 'package_name>=1.2.3'

See https://docs.astral.sh/uv/ for more information on using 'uv'.
"""

# --- Post-install instructions template ---

POST_INSTALL_MESSAGE = f"""
--- Bootstrap complete! ---

The development environment has been set up in the '{ACTIVATED_VENV_DIR}' directory,
and the project has been installed in editable mode.

To activate the project's development virtual environment, run:

  {{activate}}

To deactivate the virtual environment, run:

  deactivate

{TOOL_USAGE_INSTRUCTIONS}
"""

# --- Confirmation prompt message ---

CONFIRMATION_PROMPT_MESSAGE = f"""
This script will create a {VENV_DIR} directory in the root
of the current repository.

It will install required tools into it for development,
and install the project as an editable package into the
virtual environment.

No changes will be made to your system install of Python.

Continue? [y/n] """

# --- Global flags for output control ---

# These are defined here only for declaration purposes; they are actually set
# in main() after parsing command-line arguments.
# Changing these variables directly has no effect: Set DEFAULT_DEBUG and
# DEFAULT_QUIET instead to change the default behavior.
DEBUG: bool = False
QUIET: bool = False


# --- Version Control System (VCS) Hook Names ---

# Standard Git and Mercurial hook names. If you need custom hooks,
# you can modify these sets as needed. Only names included in these
# sets will be installed by the install_vcs_hooks() function.
GIT_HOOK_NAMES = {
    'applypatch-msg',
    'commit-msg',
    'fsmonitor-watchman',
    'post-update',
    'pre-applypatch',
    'pre-commit',
    'pre-merge-commit',
    'pre-push',
    'pre-rebase',
    'pre-receive',
    'prepare-commit-msg',
    'update',
}

HG_HOOK_NAMES = {
    'precommit',
    'commit',
    'prepush',
    'push',
    'preupdate',
    'update',
    'prechangegroup',
    'changegroup',
    'pretag',
    'tag',
}


class VCS(NamedTuple):
    """Specification for version control systems.

    :param str name: The name of the VCS (e.g., 'git', 'hg', 'none').
    """

    name: str
    repo_root: Path | None = None

    def is_git(self) -> bool:
        """Returns True if the VCS is git."""
        return self.name == 'git'

    def is_hg(self) -> bool:
        """Returns True if the VCS is Mercurial."""
        return self.name == 'hg'

    def is_none(self) -> bool:
        """Returns True if no VCS is detected."""
        return self.name == 'none'

    def __str__(self) -> str:
        return self.name


DETECTED_VCS: VCS = VCS(name='none', repo_root=Path('.'))
"""The detected version control system in use (git, hg, or none)."""


def run_post_install_steps(python_exe: Path, root_path: Path, bin_dir: Path) -> None:
    """Runs any post-installation steps required after installing tools.

    This function is called automatically after the core development tools are installed.
    It is intended as a customization point for project-specific setup tasks, such as:
    - Installing the current project in editable mode
    - Setting up pre-commit hooks
    - Installing packages from requirements.txt files
    - Any other project-specific initialization

    The example here installs the current project in editable mode with 'uv pip install -e .'.

    :param python_exe Path: The path to the Python executable within the venv.
    :param root_path Path: The path to the root of the repository.
    """
    _validate_path(python_exe, 'python_exe', exists=True)
    _validate_path(root_path, 'root_path', exists=True)

    controlled_print('--> Installing the current project and syncing all dependency groups...')
    # We use --all-groups to ensure 'tools' (tox), 'test', etc. defined in
    # pyproject.toml are installed. uv sync also handles installing the
    # project itself in editable mode.
    run_command([bin_dir / 'uv', 'sync', '--all-groups'], cwd=root_path, check=True)


class FatalBootstrapError(Exception):
    """Exception raised for fatal errors during the bootstrap process."""

    def __init__(self, message: str, error_code: int = 1) -> None:
        super().__init__(message)
        self.error_code = error_code


def _is_windows() -> bool:
    """Determines if the current platform is Windows."""
    return sys.platform == 'win32'


def _validate_string(value: str, name: str) -> None:
    """Validates that the input is a string.

    :param value str: The value to validate.
    :param name str: The name of the value (for error messages).
    :raises TypeError: If validation fails.
    """
    if not isinstance(value, str):
        raise TypeError(f'{name} must be a string')


def _validate_string_list(lst: list[str], name: str) -> None:
    """Validates that the input is a list of strings.

    :param lst list[str]: The list to validate.
    :param name str: The name of the list (for error messages).
    :raises TypeError: If validation fails.
    """
    if not isinstance(lst, list):
        raise TypeError(f'{name} must be a list')
    if not all(isinstance(item, str) for item in lst):
        raise TypeError(f'all items in {name} must be strings')


def _validate_module_list(modules: list[InstallSpec], name: str) -> None:
    """Validates that the input is a list of InstallSpec instances.

    :param modules list[InstallSpec]: The list to validate.
    :param name str: The name of the list (for error messages).
    :raises TypeError: If validation fails.
    """
    if not isinstance(modules, list):
        raise TypeError(f'{name} must be a list')
    for module in modules:
        if not isinstance(module, InstallSpec):
            raise TypeError(f'all items in {name} must be InstallSpec instances')


def _validate_command(lst: list[str | Path], name: str) -> None:
    """Validates that the input is a list of that starts with
    either a string or Path, and contains only strings for all other items.

    It must contain at least one item.

    :param lst list[str | Path]: The list to validate.
    :param name str: The name of the list (for error messages).
    :raises TypeError: If validation fails.
    """
    if not isinstance(lst, list):
        raise TypeError(f'{name} must be a list')
    if not lst:
        raise ValueError(f'{name} must not be empty')
    if not isinstance(lst[0], (str, Path)):
        raise TypeError(f'the first item in {name} must be a string or Path')
    if not all(isinstance(item, str) for item in lst[1:]):
        raise TypeError(f'all items after the first in {name} must be strings')


def _validate_boolean(value: bool, name: str) -> None:
    """Validates that the input is a boolean.

    :param value bool: The value to validate.
    :param name str: The name of the value (for error messages).
    :raises TypeError: If validation fails.
    """
    if not isinstance(value, bool):
        raise TypeError(f'{name} must be a boolean')


def _validate_kwarg_keys_are_strings(kwargs: dict, name: str) -> None:
    """Validates that all keys in the input dictionary are strings.

    :param kwargs dict: The dictionary to validate.
    :param name str: The name of the dictionary (for error messages).
    :raises TypeError: If validation fails.
    """
    if not isinstance(kwargs, dict):
        raise TypeError(f'{name} must be a dictionary')
    if not all(isinstance(k, str) for k in kwargs.keys()):
        raise TypeError(f'all keys in {name} must be strings')


def _validate_path(path: Path, name: str, exists: bool = False) -> None:
    """Validates that the input is a Path instance.

    Optionally checks that the path exists.

    :param path Path: The path to validate.
    :param name str: The name of the path (for error messages).
    :param exists bool: Whether to check that the path exists.
    :raises TypeError: If validation fails.
    :raises FileNotFoundError: If exists is True and the path does not exist.
    """
    if not isinstance(path, Path):
        raise TypeError(f'{name} must be a Path instance')
    if exists and not path.exists():
        raise FileNotFoundError(f'{name} does not exist: {path}')


def run_command(
    command: list[str | Path],
    *,
    check: bool = True,
    cwd: str | Path | None = None,
    **kwargs,  # noqa: ANN003
) -> None:
    """Helper to run a command and print its output.

    If the command is not found, or returns a non-zero exit code,
    prints an error message and exits the script.

    :param command list[str | Path]: The command to run as a list.
    :param check bool: Whether to raise an exception on non-zero exit code.
    :param cwd str | Path | None: The working directory for the command.
    :param kwargs: Additional keyword arguments to pass to subprocess.run().
    """
    _validate_command(command, 'command')
    _validate_boolean(check, 'check')
    if cwd:
        _validate_path(Path(cwd), 'cwd', exists=True)
    _validate_kwarg_keys_are_strings(kwargs, 'kwargs')

    try:
        if DEBUG:
            debug_kwargs = kwargs.copy()
            if cwd:
                debug_kwargs['cwd'] = cwd
            print(f'DEBUG: Running {command} with kwargs: {debug_kwargs}')
        # Suppress output if QUIET is True and not already overridden
        if QUIET:
            kwargs.setdefault('stdout', subprocess.DEVNULL)
            kwargs.setdefault('stderr', subprocess.DEVNULL)
        subprocess.run(command, check=check, cwd=cwd, **kwargs)
    except FileNotFoundError as e:
        print(f"Error: Command '{command[0]}' not found. Is it in your PATH?")
        raise FatalBootstrapError(f"Command '{command[0]}' not found.", error_code=1) from e
    except subprocess.CalledProcessError as e:
        print(f'Error: Command {command} failed with exit code {e.returncode}')
        raise FatalBootstrapError(
            f'Command {command} failed with exit code {e.returncode}', error_code=e.returncode
        ) from e


def controlled_print(message: str) -> None:
    """Prints a message if not in quiet mode."""
    _validate_string(message, 'message')
    if not QUIET:
        print(message)


def confirmation_prompt(message: str) -> bool:
    """Prompts the user for confirmation to proceed."""
    try:
        repo_root = get_repo_root()
        controlled_print(f'Current working directory: {os.getcwd()}')
        controlled_print(f'Repository root directory: {repo_root}')
        choice = ''
        while choice.lower().strip() not in ('y', 'yes', 'n', 'no'):
            choice = input(message)
    except KeyboardInterrupt:
        controlled_print('')
        return False

    return choice.lower().strip() in ('', 'y', 'yes')


@cache
def get_repo_root() -> Path:
    """Finds the root directory of the repository and caches the result.

    If not in a repository, prints an error message and exits.

    It tries to use 'git rev-parse --show-toplevel' first, and falls back
    to searching parent directories for a '.git' folder if the git command
    is not found.

    If a .git directory is not found, it looks for a Mercurial repository
    by searching for a '.hg' folder instead.

    :raises FatalBootstrapError: If no repository is found.
    """
    global DETECTED_VCS  # pylint: disable=global-statement
    try:
        git_root_bytes = subprocess.check_output(['git', 'rev-parse', '--show-toplevel'], stderr=subprocess.PIPE)
        root_dir = git_root_bytes.decode('utf-8').strip()
        DETECTED_VCS = VCS(name='git', repo_root=Path(root_dir))
        return Path(root_dir)
    except (FileNotFoundError, subprocess.CalledProcessError):
        # Try Mercurial CLI
        try:
            hg_root_bytes = subprocess.check_output(['hg', 'root'], stderr=subprocess.PIPE)
            root_dir = hg_root_bytes.decode('utf-8').strip()
            DETECTED_VCS = VCS(name='hg', repo_root=Path(root_dir))
            return Path(root_dir)
        except (FileNotFoundError, subprocess.CalledProcessError):
            # Fallback to directory search...
            current_dir = Path.cwd()
            for parent in [current_dir] + list(current_dir.parents):
                if (parent / '.git').is_dir():
                    DETECTED_VCS = VCS(name='git', repo_root=parent)
                    return parent

            # Check for Mercurial repository instead
            for parent in [current_dir] + list(current_dir.parents):
                if (parent / '.hg').is_dir():
                    DETECTED_VCS = VCS(name='hg', repo_root=parent)
                    return parent

            controlled_print('Error: No Git or Mercurial repository found in any parent directories.')
            raise FatalBootstrapError('No repository found.', error_code=1)  # noqa: B904


def install_vcs_hooks(repo_root: Path, forced: bool) -> None:
    """Install pre-commit and pre-push hooks for git or hg if present.

    It uses the 'hooks/' directory in the repository root
    to find the hook scripts to install to the auto-detected VCS
    metadata directory.

    :param repo_root Path: The root directory of the repository.
    :param forced bool: Whether to force overwrite of existing hooks.
    """
    _validate_path(repo_root, 'repo_root', exists=True)

    if DETECTED_VCS.is_none():
        controlled_print('No version control system detected; skipping VCS hook installation.')
        return

    hooks_dir = repo_root / 'hooks'
    if not hooks_dir.exists():
        controlled_print('No hooks/ directory found; skipping VCS hook installation.')
        return

    if DETECTED_VCS.is_git():
        _install_git_hooks(repo_root, forced)
    elif DETECTED_VCS.is_hg():
        _install_hg_hooks(repo_root, forced)
    else:
        controlled_print('Unsupported VCS for hook installation; skipping.')


def _is_valid_git_hook_name(name: str) -> bool:
    """Checks if the given name is a valid Git hook name.

    :param name str: The hook name to check.
    :return bool: True if valid, False otherwise.
    """
    return name in GIT_HOOK_NAMES


def _is_valid_hg_hook_name(name: str) -> bool:
    """Checks if the given name is a valid Mercurial hook name.

    :param name str: The hook name to check.
    :return bool: True if valid, False otherwise.
    """
    return name in HG_HOOK_NAMES


def _install_git_hooks(repo_root: Path, forced: bool) -> None:
    """Installs git hooks from the 'hooks/' directory in the repository root.

    :param repo_root Path: The root directory of the repository.
    :param forced bool: Whether to force overwrite existing hooks.
    """
    _validate_path(repo_root, 'repo_root', exists=True)

    git_dir = repo_root / '.git'
    hooks_dir = git_dir / 'hooks'
    source_hooks_dir = repo_root / 'hooks'

    if not source_hooks_dir.exists():
        controlled_print('No hooks/ directory found; skipping Git hook installation.')
        return

    if not hooks_dir.exists():
        controlled_print(f'Git hooks directory does not exist at {hooks_dir}; creating it.')
        try:
            hooks_dir.mkdir(parents=True, exist_ok=True)
        except Exception as e:
            controlled_print(f'Error: Could not create Git hooks directory at {hooks_dir}: {e}')
            controlled_print('Skipping Git hooks installation.')
            return

    try:
        for hook_file in source_hooks_dir.iterdir():
            if hook_file.is_file():
                hook_name = hook_file.name
                if not _is_valid_git_hook_name(hook_name):
                    controlled_print(f'Skipping non-standard Git hook: {hook_name}')
                    continue
                dest_hook = hooks_dir / hook_file.name
                if dest_hook.exists():
                    if forced:
                        controlled_print(f'Overwriting existing git hook: {dest_hook}')
                    else:
                        controlled_print(f'Git hook already exists, skipping: {dest_hook}')
                        continue
                try:
                    shutil.copy2(hook_file, dest_hook)
                except Exception as e:
                    controlled_print(f'Error: Could not copy hook {hook_file} to {dest_hook}: {e}')
                    controlled_print('Skipping this hook.')
                    continue
                try:
                    dest_hook.chmod(dest_hook.stat().st_mode | stat.S_IEXEC)
                except Exception:
                    controlled_print(f'Warning: Could not set executable permission for {dest_hook}')
                controlled_print(f'Installed git hook: {dest_hook}')
    except Exception as e:
        controlled_print(f'Error while installing Git hooks: {e}')
        controlled_print('Skipping Git hooks installation. Some hooks may not be installed.')


def _install_hg_hooks(repo_root: Path, forced: bool) -> None:
    """Installs Mercurial hooks from the 'hooks/' directory in the repository root.

    :param repo_root Path: The root directory of the repository.
    :param forced bool: Whether to force overwrite existing hooks.
    """
    _validate_path(repo_root, 'repo_root', exists=True)

    hg_dir = repo_root / '.hg'
    hgrc_file = hg_dir / 'hgrc'
    source_hooks_dir = repo_root / 'hooks'
    relative_hooks_path = '../hooks'

    if not source_hooks_dir.exists():
        controlled_print('No hooks/ directory found; skipping Mercurial hook installation.')
        return

    try:
        if not hgrc_file.exists():
            controlled_print(f'No hgrc file found at {hgrc_file}; creating a new one.')
            hgrc_file.touch()
    except Exception as e:
        controlled_print(f'Error: Could not create hgrc file at {hgrc_file}: {e}')
        controlled_print('Skipping Mercurial hooks installation.')
        return

    try:
        installed_hooks: set[str] = _already_installed_hg_hooks(repo_root)
        hook_entries = []
        hook_names = set()
        for hook_file in source_hooks_dir.iterdir():
            if hook_file.is_file():
                hook_name = hook_file.name
                if not _is_valid_hg_hook_name(hook_name):
                    controlled_print(f'Skipping non-standard Mercurial hook: {hook_name}')
                    continue
                if hook_name in installed_hooks and not forced:
                    controlled_print(f'Mercurial hook already configured, skipping: {hook_name}')
                    continue
                hook_entries.append(f'{hook_name} = {relative_hooks_path}/{hook_name}\n')
                hook_names.add(hook_name)
    except Exception as e:
        controlled_print(f'Error while preparing Mercurial hooks: {e}')
        controlled_print('Skipping Mercurial hooks installation.')
        return

    if not hook_entries:
        controlled_print('No Mercurial hooks to install; skipping.')
        return

    try:
        uniquifier = 1
        new_hgrc_file = hgrc_file.with_name(hgrc_file.name + f'new_{uniquifier}')
        while new_hgrc_file.exists():
            uniquifier += 1
            new_hgrc_file = hgrc_file.with_name(hgrc_file.name + f'new_{uniquifier}')

        if hook_entries:
            hgrc_content: list[str] = hgrc_file.read_text().splitlines(keepends=True)
            found_hooks_section = False
            with new_hgrc_file.open('w') as hgrc:
                in_hooks_section = False
                for line in hgrc_content:
                    # Find [hooks] section
                    stripped_line = line.strip()
                    if stripped_line.startswith('[hooks]'):
                        in_hooks_section = True
                        found_hooks_section = True
                        hgrc.write(line)
                        continue

                    # Process lines in [hooks] section
                    if in_hooks_section:
                        # end of [hooks] section
                        if stripped_line.startswith('[') and stripped_line.endswith(']'):
                            in_hooks_section = False
                            # Add any remaining hooks before leaving section
                            for name in hook_names:
                                hgrc.write(f'{name} = {relative_hooks_path}/{name}\n')

                        # A hook entry
                        elif '=' in line:
                            hook_name_key, hook_file_value = line.split('=', 1)
                            hook_name_key = hook_name_key.strip()
                            if hook_name_key in hook_names:
                                hook_names.remove(hook_name_key)
                                if forced:
                                    controlled_print(f'Overwriting existing Mercurial hook: {hook_name_key}')
                                    line = f'{hook_name_key} = {relative_hooks_path}/{hook_name_key}\n'
                                else:
                                    controlled_print(f'Mercurial hook already configured, skipping: {hook_name_key}')
                    hgrc.write(line)

                # If no [hooks] section was found, add it at the end
                if not found_hooks_section:
                    hgrc.write('[hooks]\n')
                    hgrc.writelines(hook_entries)

            # Replace original hgrc with new one (backup original)
            backup_suffix: str = '.bak.'
            backup_uniquifier: int = 1
            backup_file = hgrc_file.with_name(hgrc_file.name + f'{backup_suffix}{backup_uniquifier}')
            while backup_file.exists():
                backup_uniquifier += 1
                backup_file = hgrc_file.with_name(hgrc_file.name + f'{backup_suffix}{backup_uniquifier}')
            shutil.copy2(hgrc_file, backup_file)
            shutil.move(new_hgrc_file, hgrc_file)

    except Exception as e:
        controlled_print(f'Error while installing Mercurial hooks: {e}')
        controlled_print('Skipping Mercurial hooks installation.')
        return

    controlled_print(f'Configured Mercurial hooks in {hgrc_file}')


def _already_installed_hg_hooks(repo_root: Path) -> set[str]:
    """Returns a set of already installed Mercurial hook names from hgrc.

    :param repo_root Path: The root directory of the repository.
    :return set[str]: A set of installed hook names.
    """
    _validate_path(repo_root, 'repo_root', exists=True)

    hg_dir: Path = repo_root / '.hg'
    hgrc_file: Path = hg_dir / 'hgrc'
    installed_hooks: set[str] = set()

    if not hgrc_file.exists():
        return installed_hooks

    with hgrc_file.open('r') as hgrc:
        in_hooks_section = False
        for line in hgrc:
            line = line.strip()
            if line.startswith('[hooks]'):
                in_hooks_section = True
                continue
            if in_hooks_section:
                if line.startswith('[') and line.endswith(']'):
                    break  # End of hooks section
                if '=' in line:
                    hook_name = line.split('=')[0].strip()
                    installed_hooks.add(hook_name)

    return installed_hooks


def path_to_venv_python(venv_dir: Path) -> Path:
    """Returns the path to the Python executable within the virtual environment.

    :param venv_dir Path: The directory of the virtual environment.
    :param is_windows bool: Whether the platform is Windows.
    :return: The path to the Python executable.
    """
    _validate_path(venv_dir, 'venv_dir', exists=False)
    is_windows = _is_windows()
    bin_dir = venv_dir / ('Scripts' if is_windows else 'bin')
    python_exe = bin_dir / ('python.exe' if is_windows else 'python')
    return python_exe


@cache
def pip_module_is_available(python_exe: Path) -> bool:
    """Checks if 'pip' is available in the given Python executable.

    :param python_exe Path: The path to the Python executable.
    :return: True if 'pip' is available, False otherwise.
    """
    _validate_path(python_exe, 'python_exe', exists=True)

    stdout = subprocess.PIPE if not QUIET else subprocess.DEVNULL
    stderr = subprocess.PIPE if not QUIET else subprocess.DEVNULL
    try:
        if DEBUG:
            controlled_print(f"DEBUG: Running '{python_exe} -m pip --version' to check for pip availability")
        subprocess.run([python_exe, '-m', 'pip', '--version'], check=True, stdout=stdout, stderr=stderr)
        return True
    except (FileNotFoundError, subprocess.CalledProcessError):
        return False


def create_virtual_environment(venv_dir: Path, python_exe: Path) -> None:
    """
    Creates a virtual environment at the specified directory.
    If the directory already exists, it skips creation.
    :param venv_dir Path: The directory to create the virtual environment in.
    :param python_exe Path: The path to the Python executable within the venv.
    """
    _validate_path(venv_dir, 'venv_dir', exists=False)
    _validate_path(python_exe, 'python_exe', exists=False)

    if not venv_dir.exists():
        controlled_print(f"Creating temporary virtual environment in '{venv_dir}'...")
        create_venv(venv_dir, with_pip=True)
        controlled_print('---> Ensuring pip CLI script is installed in the virtual environment...')
        run_command([python_exe, '-m', 'ensurepip', '--upgrade'])

        controlled_print('---> Upgrading pip in the virtual environment to latest version...')
        if not pip_module_is_available(python_exe):
            pip_path = venv_dir / 'Scripts' / 'pip.exe' if _is_windows() else venv_dir / 'bin' / 'pip'
            if not pip_path.exists():
                controlled_print("Error: 'pip' is not available in the virtual environment after ensurepip.")
                controlled_print('Please check your Python installation.')
                raise FatalBootstrapError("'pip' not available in virtual environment.", error_code=1)
            run_command([pip_path, 'install', '--upgrade', 'pip'])
        else:
            run_command([python_exe, '-m', 'pip', 'install', '--upgrade', 'pip', '--require-virtualenv'])
    else:
        controlled_print(f"Virtual environment '{venv_dir}' already exists. Skipping creation.")


def remove_virtual_environment(venv_dir: Path, quiet: bool = False) -> None:
    """Removes the temporary virtual environment directory.

    :param venv_dir Path: The directory of the virtual environment to remove.
    """
    if not quiet:
        controlled_print(f"Removing temporary virtual environment at '{venv_dir}'...")
    if venv_dir.exists():
        if sys.version_info >= (3, 12):
            shutil.rmtree(venv_dir, onexc=_remove_readonly)
        else:  # error handler deprecated in 3.12+
            shutil.rmtree(venv_dir, onerror=_remove_readonly)


def _remove_readonly(func: Callable[[str | Path], None], path: str | Path, _: Any) -> None:
    "Clear the readonly bit and reattempt the removal"
    os.chmod(path, stat.S_IWRITE)
    func(path)


def install_tools(python_exe: Path, modules: list[InstallSpec]) -> None:
    """Installs core development tools into the virtual environment.

    If 'uv' is specified in the modules, it is bootstrapped with pip
    and used to install all the other modules; otherwise, the installation
    falls back to 'pip' for all modules.

    :param python_exe Path: The path to the Python executable within the venv.
    :param modules list[InstallSpec]: A list of InstallSpec objects to install.
    """
    _validate_path(python_exe, 'python_exe', exists=True)
    _validate_module_list(modules, 'modules')

    if not modules:
        return

    controlled_print('Installing/updating core development tools...')
    using_uv = any(mod.name == 'uv' for mod in modules)
    if using_uv:
        install_with_uv(python_exe, modules)
    else:
        install_with_pip(python_exe, modules)


def install_with_uv(python_exe: Path, modules: list[InstallSpec]) -> None:
    """Installs 'uv' using pip, then uses 'uv' to install the specified modules.

    :param python_exe Path: The path to the Python executable within the venv.
    :param modules list[InstallSpec]: A list of InstallSpec objects to install.
    """
    _validate_path(python_exe, 'python_exe', exists=True)
    _validate_module_list(modules, 'modules')

    uv_module: InstallSpec = [mod for mod in modules if mod.name == 'uv'][0]
    other_modules: list[InstallSpec] = [mod for mod in modules if mod.name != 'uv']

    bootstrap_message = f"--> Bootstrapping 'uv' using 'pip': {uv_module}, {uv_module.version or 'latest'}"
    install_with_pip(python_exe, [uv_module], message=bootstrap_message)

    if not other_modules:
        return

    controlled_print("--> Installing remaining modules using 'uv pip'")
    command = _build_install_command([python_exe, '-m', 'uv', 'pip'], other_modules)
    run_command(command)


def install_with_pip(python_exe: Path, modules: list[InstallSpec], message: str = '') -> None:
    """Installs the specified modules using 'pip'.

    :param python_exe Path: The path to the Python executable within the venv.
    :param modules: A list of InstallSpec objects to install.
    :param message str: An optional message to print before installation.
    """
    _validate_path(python_exe, 'python_exe', exists=True)
    _validate_module_list(modules, 'modules')
    _validate_string(message, 'message')

    if message:
        controlled_print(message)
    else:
        controlled_print("--> Installing modules using 'pip'")
    command = _build_install_command([python_exe, '-m', 'pip', '--require-virtualenv'], modules)
    run_command(command)


def _build_install_command(base_command: list[str | Path], modules: list[InstallSpec]) -> list[str | Path]:
    """Builds a complete installation command list for either 'pip' or 'uv pip'.

    :param base_command list[str | Path]: The base command to start with (e.g., pip or uv pip).
    :param modules list[InstallSpec]: A list of InstallSpec objects to install.
    :return list[str | Path]: The complete command list to run.
    """
    _validate_command(base_command, 'base_command')
    _validate_module_list(modules, 'modules')

    command = base_command + ['install', '-U']
    for module in modules:
        spec_str = module.name
        if module.extras:
            spec_str += module.extras
        if module.version:
            spec_str += module.version
        command.append(spec_str)
    return command


def print_instructions(template: str) -> None:
    """Prints instructions to the user on how to activate the virtual environment
    and use the installed tools.

    :param template str: The instructions template to use.
    """
    _validate_string(template, 'template')

    activate_script = f"""
Unix/macOS/Linux:

    source {ACTIVATED_VENV_DIR}/bin/activate

If you use fish:

    source {ACTIVATED_VENV_DIR}/bin/activate.fish
"""

    if _is_windows():
        activate_script = f"""
Windows Command Prompt:

    {ACTIVATED_VENV_DIR}\\Scripts\\activate.bat

PowerShell:

    {ACTIVATED_VENV_DIR}\\Scripts\\Activate.ps1

Note: For PowerShell, you may need to adjust your execution policy to run scripts.
You can do this by running PowerShell and executing:

    Set-ExecutionPolicy -Scope CurrentUser RemoteSigned
"""
    instructions = template.format(activate=activate_script)
    controlled_print(instructions)


def parse_arguments() -> argparse.Namespace:
    """Parses command-line arguments."""
    arg_parser = argparse.ArgumentParser(
        description='Bootstrap the development environment by creating a '
        'virtual environment and installing required tools.'
    )
    arg_parser.add_argument(
        '--yes', '-y', action='store_true', help='Automatically confirm and proceed without prompting.'
    )

    debug_group = arg_parser.add_mutually_exclusive_group()
    debug_group.add_argument('--debug', dest='debug', action='store_true', help='Enable debug output.')
    debug_group.add_argument('--no-debug', dest='debug', action='store_false', help='Disable debug output.')

    # Mutually exclusive group for verbosity
    verbosity_group = arg_parser.add_mutually_exclusive_group()
    verbosity_group.add_argument(
        '-q',
        '--quiet',
        '--ci',
        dest='quiet',
        action='store_true',
        help='Suppress non-error output. This is useful for CI environments.',
    )
    verbosity_group.add_argument(
        '-v', '--verbose', dest='quiet', action='store_false', help='Enable verbose output (default).'
    )

    # Forced hook installation option
    arg_parser.add_argument(
        '--force-hooks', action='store_true', help='Force overwrite of existing VCS hooks during installation.'
    )
    arg_parser.set_defaults(quiet=DEFAULT_QUIET, debug=DEFAULT_DEBUG)

    return arg_parser.parse_args()


def main() -> None:
    """
    Checks for required development tools and bootstraps a local virtual
    environment with them if necessary.
    """
    args = parse_arguments()
    global DEBUG, QUIET  # pylint: disable=global-statement
    DEBUG = args.debug
    QUIET = args.quiet

    if QUIET and not args.yes:
        print('Note: You can use --yes/-y to skip confirmation prompts.')
    if not args.yes and not confirmation_prompt(CONFIRMATION_PROMPT_MESSAGE):
        print('Aborted by user.')
        sys.exit(0)

    try:
        repo_root = get_repo_root()

        controlled_print(f'--- Bootstrapping development environment (in {repo_root}) ---')

        venv_dir = repo_root / VENV_DIR
    except FatalBootstrapError as e:
        controlled_print(f'Fatal error during bootstrap: {e}')
        sys.exit(e.error_code)

    try:
        python_exe = path_to_venv_python(venv_dir)
        create_virtual_environment(venv_dir, python_exe)
        install_tools(python_exe, BOOTSTRAP_MODULES)
        install_vcs_hooks(repo_root, forced=args.force_hooks)

        bin_dir = venv_dir / ('Scripts' if _is_windows() else 'bin')
        run_post_install_steps(python_exe=python_exe, root_path=repo_root, bin_dir=bin_dir)
        if REMOVE_BOOTSTRAP_VENV_ON_EXIT:
            remove_virtual_environment(venv_dir)
        print_instructions(POST_INSTALL_MESSAGE)

    except KeyboardInterrupt:
        remove_virtual_environment(venv_dir, quiet=True)
        controlled_print('')
        controlled_print('Aborted by user.')
        sys.exit(2)

    except FatalBootstrapError as e:
        remove_virtual_environment(venv_dir, quiet=True)
        controlled_print(f'Fatal error during bootstrap: {e}')
        sys.exit(e.error_code)

    except Exception as e:
        remove_virtual_environment(venv_dir, quiet=True)
        controlled_print(f'Fatal error during bootstrap: {e}')
        sys.exit(1)


if __name__ == '__main__':
    main()



"""Tests for :mod:`autopypath._config_py_path._config_py_path`."""

import logging
import os
import sys
from collections.abc import Sequence
from pathlib import Path

import pytest

from autopypath import AutopypathError
from autopypath import _defaults as defaults
from autopypath._config_py_path._config_py_path import _EMPTY_AUTOPYPATH_CONFIG, _NON_RESOLVABLE_SYS_PATH, _ConfigPyPath
from autopypath._types import _NoPath

_ORIGINAL_SYS_PATH: list[str] = sys.path.copy()
_ORIGINAL_NAME: str = __name__


def setup_function() -> None:
    """Setup function to reset sys.path before each test."""
    sys.path = _ORIGINAL_SYS_PATH.copy()


def teardown_function() -> None:
    """Teardown function to reset sys.path after each test."""
    sys.path = _ORIGINAL_SYS_PATH.copy()


def test_empty_autopypath_config() -> None:
    """Tests the special empty AutopypathConfig instance."""
    assert isinstance(_EMPTY_AUTOPYPATH_CONFIG.toml_filepath, _NoPath), (
        'EMPTY_AUTO_001 _EMPTY_AUTOPYPATH_CONFIG.toml_filepath should be a _NoPath instance'
    )
    assert _EMPTY_AUTOPYPATH_CONFIG.repo_markers is None, (
        'EMPTY_AUTO_002 _EMPTY_AUTOPYPATH_CONFIG.repo_markers should be None'
    )
    assert _EMPTY_AUTOPYPATH_CONFIG.paths is None, 'EMPTY_AUTO_003 _EMPTY_AUTOPYPATH_CONFIG.paths should be None'
    assert _EMPTY_AUTOPYPATH_CONFIG.load_strategy is None, (
        'EMPTY_AUTO_004 _EMPTY_AUTOPYPATH_CONFIG.load_strategy should be None'
    )
    assert _EMPTY_AUTOPYPATH_CONFIG.path_resolution_order is None, (
        'EMPTY_AUTO_005 _EMPTY_AUTOPYPATH_CONFIG.path_resolution_order should be None'
    )


def test_configured_autopypath_config(tmp_path: Path) -> None:
    """Tests that _ConfigPyPath.autopypath_config returns the correct configuration
    when an autopypath.toml file exists in the root of the repository.

    We want to test various combinations of existing and non-existing paths
    specified in the autopypath.toml file.

    'src' exists, 'tests' does not exist.

    Uses a temporary directory to simulate a repository with an autopypath.toml file.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    src_path = root_path / 'src'
    src_path.mkdir()
    git_path = root_path / '.git'
    git_path.mkdir()
    autopypath_path = root_path / 'autopypath.toml'
    # Note: only "src" exists, "tests" does not exist
    autopypath_path.write_text("""
[tool.autopypath]
load_strategy = "replace"
path_resolution_order = ["manual", "autopypath", "pyproject"]
repo_markers = {".git" = "dir", "setup.py" = "file"}
paths=["src", "tests"]
""")
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        dry_run=True,
    )
    autopypath_config = config.autopypath_config
    assert autopypath_config.load_strategy == 'replace', (
        'AUTOPYPATH_CONFIGURED_001 autopypath_config.load_strategy should be LoadStrategy.REPLACE'
    )
    assert autopypath_config.path_resolution_order == (
        'manual',
        'autopypath',
        'pyproject',
    ), (
        'AUTOPYPATH_CONFIGURED_002 autopypath_config.path_resolution_order should match the configured '
        ' path resolution order: ["manual", "autopypath", "pyproject"]'
    )
    assert autopypath_config.repo_markers == {'.git': 'dir', 'setup.py': 'file'}, (
        'AUTOPYPATH_CONFIGURED_003 autopypath_config.repo_markers should match the configured repo markers'
    )
    if isinstance(autopypath_config.paths, Sequence):
        assert len(autopypath_config.paths) == 2, (
            'AUTOPYPATH_CONFIGURED_004 autopypath_config.paths should '
            'have length 2 because src and tests were configured'
        )
        assert str(autopypath_config.paths[0].name) == 'src', (
            'AUTOPYPATH_CONFIGURED_005 autopypath_config.paths should match the configured paths'
        )
        assert autopypath_config.paths[0].is_dir(), (
            'AUTOPYPATH_CONFIGURED_006 autopypath_config.paths[0] should be match the path to src directory'
        )
        assert src_path.resolve() == autopypath_config.paths[0].resolve(), (
            'AUTOPYPATH_CONFIGURED_007 autopypath_config.paths[0] should resolve to the src directory path'
        )
    else:
        pytest.fail('AUTOPYPATH_CONFIGURED_008 autopypath_config.paths should be a list')


def test_default_config(tmp_path: Path) -> None:
    """Tests that _ConfigPyPath.autopypath_config returns the special empty config when appropriate.
    Uses a temporary directory to simulate a repository without any autopypath configuration
    and a pyproject.toml file to mark the root.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    pyproject_path = root_path / 'pyproject.toml'
    pyproject_path.write_text("""
[tool.some_other_tool]
""")

    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        dry_run=True,
    )

    default_config = config.default_config
    assert default_config.load_strategy == defaults._LOAD_STRATEGY, (
        'DEFAULT_CONFIG_001 default_config.load_strategy should match defaults._LOAD_STRATEGY'
    )
    assert default_config.path_resolution_order == defaults._PATH_RESOLUTION_ORDER, (
        'DEFAULT_CONFIG_002 default_config.path_resolution_order should match defaults._PATH_RESOLUTION_ORDER'
    )
    assert default_config.repo_markers == defaults._REPO_MARKERS, (
        'DEFAULT_CONFIG_003 default_config.repo_markers should match defaults._REPO_MARKERS'
    )
    assert default_config.paths == defaults._PATHS


def test_no_pyproject_toml_file(tmp_path: Path) -> None:
    """Tests that _ConfigPyPath.pyproject_config returns an empty config when
    no pyproject.toml configuration file exists

    Uses a temporary directory to simulate a repository with an empty autopypath config in pyproject.toml.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    git_path = root_path / '.git'
    git_path.mkdir()

    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        dry_run=True,
    )

    pyproject_config = config.pyproject_config
    assert pyproject_config.load_strategy is None, 'NO_PYPROJECT_FILE_001 pyproject_config.load_strategy should be None'
    assert pyproject_config.path_resolution_order is None, (
        'NO_PYPROJECT_FILE_002 pyproject_config.path_resolution_order should be None'
    )
    assert pyproject_config.repo_markers is None, 'NO_PYPROJECT_FILE_003 pyproject_config.repo_markers should be None'
    assert pyproject_config.paths is None, 'NO_PYPROJECT_FILE_004 pyproject_config.paths should be None'


def test_empty_pyproject_config(tmp_path: Path) -> None:
    """Tests that _ConfigPyPath.pyproject_config returns an empty config when pyproject.toml
    exists but has no autopypath configuration.

    Uses a temporary directory to simulate a repository with an empty autopypath config in pyproject.toml.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    pyproject_path = root_path / 'pyproject.toml'
    pyproject_path.write_text("""
[tool.some_other_tool]
""")
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        dry_run=True,
    )

    pyproject_config = config.pyproject_config
    assert pyproject_config.load_strategy is None, (
        'EMPTY_PYPROJECT_CONFIG_001 pyproject_config.load_strategy should be None'
    )
    assert pyproject_config.path_resolution_order is None, (
        'EMPTY_PYPROJECT_CONFIG_002 pyproject_config.path_resolution_order should be None'
    )
    assert pyproject_config.repo_markers is None, (
        'EMPTY_PYPROJECT_CONFIG_003 pyproject_config.repo_markers should be None'
    )
    assert pyproject_config.paths is None, 'EMPTY_PYPROJECT_CONFIG_004 pyproject_config.paths should be None'


def test_configured_pyproject_config(tmp_path: Path) -> None:
    """Tests that _ConfigPyPath.pyproject_config returns the correct configuration
    when a pyproject.toml file exists in the root of the repository.

    We want to test various combinations of existing and non-existing paths
    specified in the autopypath.toml file.

    'src' exists, 'tests' does not exist.

    Uses a temporary directory to simulate a repository with an autopypath.toml file.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    src_path = root_path / 'src'
    src_path.mkdir()
    git_path = root_path / '.git'
    git_path.mkdir()
    pyproject_path = root_path / 'pyproject.toml'
    # Note: only "src" exists, "tests" does not exist
    pyproject_path.write_text("""
[tool.autopypath]
load_strategy = "prepend_highest_priority"
path_resolution_order = ["manual", "autopypath", "pyproject"]
repo_markers = {".git" = "dir", "setup.py" = "file"}
paths=["src", "tests"]
""")
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        dry_run=True,
    )
    pyproject_config = config.pyproject_config
    assert pyproject_config.load_strategy == 'prepend_highest_priority', (
        'PYPROJECT_CONFIGURED_001 pyproject_config.load_strategy should be LoadStrategy.PREPEND_HIGHEST_PRIORITY'
    )
    assert pyproject_config.path_resolution_order == (
        'manual',
        'autopypath',
        'pyproject',
    ), (
        'PYPROJECT_CONFIGURED_002 pyproject_config.path_resolution_order should match the configured '
        ' path resolution order: ["manual", "autopypath", "pyproject"]'
    )
    assert pyproject_config.repo_markers == {'.git': 'dir', 'setup.py': 'file'}, (
        'PYPROJECT_CONFIGURED_003 pyproject_config.repo_markers should match the configured repo markers'
    )
    if isinstance(pyproject_config.paths, Sequence):
        assert len(pyproject_config.paths) == 2, (
            'PYPROJECT_CONFIGURED_004 pyproject_config.paths should have length 2 because src and tests were configured'
        )
        assert str(pyproject_config.paths[0].name) == 'src', (
            'PYPROJECT_CONFIGURED_005 pyproject_config.paths should match the configured paths'
        )
        assert str(pyproject_config.paths[1].name) == 'tests', (
            'PYPROJECT_CONFIGURED_006 pyproject_config.paths should match the configured paths'
        )
        assert pyproject_config.paths[0].is_dir(), (
            'PYPROJECT_CONFIGURED_007 pyproject_config.paths[0] should be match the path to src directory'
        )
        assert src_path.resolve() == pyproject_config.paths[0].resolve(), (
            'PYPROJECT_CONFIGURED_008 pyproject_config.paths[0] should resolve to the src directory path'
        )
    else:
        pytest.fail('PYPROJECT_CONFIGURED_009 pyproject_config.paths should be a list')


def test_no_autopypath_toml_file(tmp_path: Path) -> None:
    """Tests that _ConfigPyPath.autopypath_config returns the special empty config when
    no autopypath.toml configuration file exists

    Uses a temporary directory to simulate a repository with an empty autopypath config in pyproject.toml.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    pyproject_path = root_path / 'pyproject.toml'
    pyproject_path.write_text("""
[tool.some_other_tool]
""")
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        dry_run=True,
    )

    autopypath_config = config.autopypath_config
    assert autopypath_config.load_strategy is None, (
        'NO_AUTOPYPATH_FILE_001 autopypath_config.load_strategy should be None'
    )
    assert autopypath_config.path_resolution_order is None, (
        'NO_AUTOPYPATH_FILE_002 autopypath_config.path_resolution_order should be None'
    )
    assert autopypath_config.repo_markers is None, (
        'NO_AUTOPYPATH_FILE_003 autopypath_config.repo_markers should be None'
    )
    assert autopypath_config.paths is None, 'NO_AUTOPYPATH_FILE_004 autopypath_config.paths should be None'


def test_manual_config(tmp_path: Path) -> None:
    """Tests that _ConfigPyPath.manual_config returns an empty config when no manual
    configuration is provided.

    Uses a temporary directory to simulate a repository without any manual configuration.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    hg_path = root_path / '.hg'
    hg_path.mkdir()

    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        dry_run=True,
        load_strategy='prepend',
        path_resolution_order=['manual', 'autopypath'],
        paths=['src', 'tests'],
        repo_markers={'.hg': 'dir'},
    )

    manual_config = config.manual_config
    assert manual_config.load_strategy, 'MANUAL_CONFIG_001 manual_config.load_strategy should not be None'
    assert manual_config.path_resolution_order, (
        'MANUAL_CONFIG_002 manual_config.path_resolution_order should not be None'
    )
    assert isinstance(manual_config.paths, Sequence), 'MANUAL_CONFIG_003 manual_config.paths should be a Sequence'
    assert manual_config.repo_markers == {'.hg': 'dir'}, (
        'MANUAL_CONFIG_004 manual_config.repo_markers should match the provided manual configuration'
    )

    if isinstance(manual_config.paths, Sequence):
        assert len(manual_config.paths) == 2, (
            'MANUAL_CONFIG_005 manual_config.paths should have length 2 because src and tests were provided'
        )
        assert str(manual_config.paths[0].name) == 'src', (
            'MANUAL_CONFIG_006 manual_config.paths should match the provided manual configuration paths'
        )
        assert str(manual_config.paths[1].name) == 'tests', (
            'MANUAL_CONFIG_007 manual_config.paths should match the provided manual configuration paths'
        )
    else:
        pytest.fail('MANUAL_CONFIG_008 manual_config.paths should be a list')


def test_replace_strategy_live(tmp_path: Path) -> None:
    """
    Performs live tests of the 'replace' load strategy.

    Default sys.path is preserved and restored after the test.

    :param Path tmp_path: Path to a temporary directory for testing.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    context_file = root_path / 'some_file.txt'
    context_file.write_text('Just a test file.')
    vcs_path = root_path / '.vcs'
    vcs_path.mkdir()

    # src and tests do not exist - expect RuntimeError
    sys_path_before: list[str] = sys.path.copy()
    try:
        _ConfigPyPath(
            context_file=root_path / 'some_file.txt',
            load_strategy='replace',
            path_resolution_order=['manual'],
            paths=['src', 'tests'],
            repo_markers={'.vcs': 'dir'},
        )
        pytest.fail('REPLACE_STRATEGY_LIVE_001 Expected AutopypathError because no paths exist to replace sys.path')

    except AutopypathError:
        pass  # Expected because paths do not exist at all
    finally:
        sys.path = sys_path_before

    # Create the 'src' directory so one of the paths exists
    try:
        src_path = root_path / 'src'
        src_path.mkdir()

        _ConfigPyPath(
            context_file=root_path / 'some_file.txt',
            load_strategy='replace',
            path_resolution_order=['manual'],
            paths=['src', 'tests'],
            repo_markers={'.vcs': 'dir'},
        )

        assert len(sys.path) == 1, (
            'REPLACE_STRATEGY_LIVE_002 sys.path should have length 1 after replace strategy is applied'
        )
        assert sys.path[0] == str(src_path.resolve()), (
            'REPLACE_STRATEGY_LIVE_003 sys.path[0] should be the src directory path'
        )
    finally:
        sys.path = sys_path_before


def test_prepend_strategy_live(tmp_path: Path) -> None:
    """
    Performs live tests of the 'prepend' load strategy.

    Default sys.path is preserved and restored after the test.

    :param Path tmp_path: Path to a temporary directory for testing.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    context_file = root_path / 'some_file.txt'
    context_file.write_text('Just a test file.')
    vcs_path = root_path / '.vcs'
    vcs_path.mkdir()

    # src and tests do not exist - expect no changes to sys.path
    sys_path_before: list[str] = sys.path.copy()
    try:
        _ConfigPyPath(
            context_file=root_path / 'some_file.txt',
            load_strategy='prepend',
            path_resolution_order=['manual'],
            paths=['src', 'tests'],
            repo_markers={'.vcs': 'dir'},
        )

        assert sys.path == sys_path_before, (
            'PREPEND_STRATEGY_LIVE_001 sys.path should be unchanged because no valid paths exist to prepend'
        )
    finally:
        sys.path = sys_path_before

    # Create the 'src' directory so one of the paths exists
    try:
        src_path = root_path / 'src'
        src_path.mkdir()

        _ConfigPyPath(
            context_file=root_path / 'some_file.txt',
            load_strategy='prepend',
            path_resolution_order=['manual'],
            paths=['src', 'tests'],
            repo_markers={'.vcs': 'dir'},
        )

        assert len(sys.path) == len(sys_path_before) + 1, (
            'PREPEND_STRATEGY_LIVE_002 sys.path should have one additional entry after prepend strategy is applied'
        )
        assert sys.path[0] == str(src_path.resolve()), (
            'PREPEND_STRATEGY_LIVE_003 sys.path[0] should be the src directory path'
        )
        assert sys.path[1:] == sys_path_before, (
            'PREPEND_STRATEGY_LIVE_004 sys.path entries after the first should be unchanged'
        )
    finally:
        sys.path = sys_path_before


def test_prepend_highest_priority_strategy_live(tmp_path: Path) -> None:
    """
    Performs live tests of the 'prepend_highest_priority' load strategy.

    Default sys.path is preserved and restored after the test.

    :param Path tmp_path: Path to a temporary directory for testing.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    context_file = root_path / 'some_file.txt'
    context_file.write_text('Just a test file.')
    vcs_path = root_path / '.vcs'
    vcs_path.mkdir()

    # src and tests do not exist - expect no changes to sys.path
    sys_path_before: list[str] = sys.path.copy()
    try:
        _ConfigPyPath(
            context_file=root_path / 'some_file.txt',
            load_strategy='prepend_highest_priority',
            path_resolution_order=['manual'],
            paths=['src', 'tests'],
            repo_markers={'.vcs': 'dir'},
        )

        assert sys.path == sys_path_before, (
            'PREPEND_HIGHEST_PRIORITY_STRATEGY_LIVE_001 sys.path should be '
            'unchanged because no valid paths exist to prepend'
        )
    finally:
        sys.path = sys_path_before

    assert sys.path == sys_path_before, (
        'PREPEND_HIGHEST_PRIORITY_STRATEGY_LIVE_002 sys.path should be have been restored to its original state'
    )
    # Create the 'src' directory so one of the paths exists
    try:
        src_path = root_path / 'src'
        src_path.mkdir()

        _ConfigPyPath(
            context_file=root_path / 'some_file.txt',
            load_strategy='prepend_highest_priority',
            path_resolution_order=['manual'],
            paths=['src', 'tests'],
            repo_markers={'.vcs': 'dir'},
        )

        assert len(sys.path) == len(sys_path_before) + 1, (
            'PREPEND_HIGHEST_PRIORITY_STRATEGY_LIVE_003 sys.path should have '
            'one additional entry after prepend_highest_priority strategy is applied'
            f': {sys.path}'
        )
        assert sys.path[0] == str(src_path.resolve()), (
            'PREPEND_HIGHEST_PRIORITY_STRATEGY_LIVE_004 sys.path[0] should be the src directory path'
        )
        assert sys.path[1:] == sys_path_before, (
            'PREPEND_HIGHEST_PRIORITY_STRATEGY_LIVE_005 sys.path entries after the first should be unchanged'
        )
    finally:
        sys.path = sys_path_before

    assert sys.path == sys_path_before, (
        'PREPEND_HIGHEST_PRIORITY_STRATEGY_LIVE_006 sys.path should be have been restored to its original state'
    )

    # Now create a autopypath.toml file that specifies a lower precedence paths
    # setting of ['tests'] rather than ['src', 'tests'] from the manual config
    # and create the 'tests' directory so that path exists
    try:
        autopypath_path = root_path / 'autopypath.toml'
        autopypath_path.write_text("""
[tool.autopypath]
load_strategy = "prepend_highest_priority"
path_resolution_order = ["autopypath", "manual"]
repo_markers = {".vcs" = "dir"}
paths=["tests"]
""")
        tests_path = root_path / 'tests'
        tests_path.mkdir()

        _ConfigPyPath(
            context_file=root_path / 'some_file.txt',
            load_strategy='prepend_highest_priority',
            path_resolution_order=['manual', 'autopypath'],
            paths=['src'],
            repo_markers={'.vcs': 'dir'},
        )
        # autopypath and manual configs both have valid paths, but manual has
        # the higher priority so only 'src' should be added to sys.path
        assert len(sys.path) == len(sys_path_before) + 1, (
            'PREPEND_HIGHEST_PRIORITY_STRATEGY_LIVE_007 sys.path should have one'
            'additional entry after prepend_highest_priority strategy is applied '
            'with autopypath.toml '
        )
        assert sys.path[0] == str(src_path.resolve()), (
            'PREPEND_HIGHEST_PRIORITY_STRATEGY_LIVE_008 sys.path[0] should be the src directory path'
        )
        assert sys.path[1:] == sys_path_before, (
            'PREPEND_HIGHEST_PRIORITY_STRATEGY_LIVE_009 sys.path entries after the first should be unchanged'
        )
    finally:
        sys.path = sys_path_before


def test_strict(tmp_path: Path, caplog: pytest.LogCaptureFixture) -> None:
    """Tests that _ConfigPyPath raises AutopypathError in strict mode
    when no valid paths are found to add to sys.path for non-replace
    load strategies.

    Uses a temporary directory to simulate a repository with an autopypath.toml file.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    git_path = root_path / '.git'
    git_path.mkdir()
    autopypath_path = root_path / 'autopypath.toml'
    # Note: both "src" and "tests" do not exist
    autopypath_path.write_text("""
[tool.autopypath]
load_strategy = "prepend"
path_resolution_order = ["manual", "autopypath", "pyproject"]
repo_markers = {".git" = "dir"}
paths=["src", "tests"]
""")
    try:
        _ConfigPyPath(
            context_file=root_path / 'some_file.txt',
            dry_run=True,
            strict=True,
        )
        pytest.fail('STRICT_001 Expected AutopypathError because no paths exist and strict mode is enabled')
    except AutopypathError:
        pass  # Expected because paths do not exist

    # Now test that a warning is logged for non-strict mode
    caplog.clear()
    _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        dry_run=True,
        strict=False,
    )
    warning_messages = [record.message for record in caplog.records if record.levelno == logging.WARNING]
    assert any(
        'autopypath: No valid paths to add to sys.path after processing.' in message for message in warning_messages
    ), 'STRICT_002 Expected a warning about no valid paths to add to sys.path in non-strict mode'


def test_symlinked_paths(tmp_path: Path) -> None:
    """Test Config with symlinked paths."""
    if os.name == 'nt':
        pytest.skip('SYMLINKED_PATHS tests are skipped on Windows due to symlink permission issues.')
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    pyproject_path = root_path / 'pyproject.toml'
    pyproject_path.write_text("""
[tool.autopypath]
paths = ["symlinked_path"]
""")
    symlink_target = root_path / 'actual_path'
    symlink_target.mkdir()
    symlink_path = root_path / 'symlinked_path'
    symlink_path.symlink_to(symlink_target, target_is_directory=True)

    try:
        sys.path = _ORIGINAL_SYS_PATH.copy()
        _ConfigPyPath(
            context_file=root_path / 'some_file.txt',
        )
        new_path_entry = str((root_path / 'symlinked_path').resolve())
        assert new_path_entry in sys.path, 'SYMLINKED_PATHS_001 The resolved symlinked path should be in sys.path'
    finally:
        sys.path = _ORIGINAL_SYS_PATH.copy()

    # remove the target directory and test that the symlink is ignored
    symlink_target.rmdir()
    try:
        sys.path = _ORIGINAL_SYS_PATH.copy()
        _ConfigPyPath(
            context_file=root_path / 'some_file.txt',
        )
        new_path_entry = str((root_path / 'symlinked_path').resolve())
        assert new_path_entry not in sys.path, (
            'SYMLINKED_PATHS_002 The symlinked path should be ignored since the target does not exist'
        )
    finally:
        sys.path = _ORIGINAL_SYS_PATH.copy()


def test_non_resolvable_sys_path_entry(tmp_path: Path, caplog: pytest.LogCaptureFixture) -> None:
    """Test that non-resolvable sys.path entries are handled gracefully."""

    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    git_path = root_path / '.git'
    git_path.mkdir()
    src_path = root_path / 'src'
    src_path.mkdir()

    bogus_path = '::nota\0path::'
    try:
        sys.path = _ORIGINAL_SYS_PATH.copy() + [bogus_path]
        caplog.clear()
        _ConfigPyPath(
            context_file=root_path / 'some_file.txt',
            log_level=logging.DEBUG,
        )
        messages = [record.message for record in caplog.records]
        if not any(message.startswith(_NON_RESOLVABLE_SYS_PATH) for message in messages):
            pytest.fail('NON_RESOLVABLE_SYS_PATH_001 Expected a log message about the non-resolvable sys.path entry')

    finally:
        sys.path = _ORIGINAL_SYS_PATH.copy()


def test_non_file_autopypath_toml(tmp_path: Path) -> None:
    """Test that a non-file autopypath.toml is handled gracefully."""

    root_path = tmp_path / 'repo'
    root_path.mkdir()
    context_file = root_path / 'some_file.txt'
    context_file.write_text('Just a test file.')
    git_path = root_path / '.git'
    git_path.mkdir()
    autopypath_path = root_path / 'autopypath.toml'
    autopypath_path.mkdir()  # Create a directory instead of a file

    try:
        sys.path = _ORIGINAL_SYS_PATH.copy()
        config = _ConfigPyPath(
            context_file=context_file,
        )
        autopypath_config = config.autopypath_config
        assert autopypath_config == _EMPTY_AUTOPYPATH_CONFIG, (
            'NON_FILE_AUTOPYPATH_TOML_001 autopypath_config should '
            'be the empty config when autopypath.toml is not a file'
        )
    finally:
        sys.path = _ORIGINAL_SYS_PATH.copy()

    # Now with strict
    try:
        sys.path = _ORIGINAL_SYS_PATH.copy()
        _ConfigPyPath(
            context_file=context_file,
            strict=True,
        )
        pytest.fail(
            'NON_FILE_AUTOPYPATH_TOML_002 Expected AutopypathError because autopypath.toml is not a file in strict mode'
        )
    except AutopypathError:
        pass
    finally:
        sys.path = _ORIGINAL_SYS_PATH.copy()


def test_no_repo(tmp_path: Path) -> None:
    """Test that a AutopypathError is raised when no repository markers are found."""
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    context_file = root_path / 'some_file.txt'
    context_file.write_text('Just a test file.')

    try:
        _ConfigPyPath(
            context_file=context_file,
            dry_run=True,
            repo_markers={'.nonexistent_vcs': 'dir'},
        )
        pytest.fail('NO_REPO_001 Expected AutopypathError because no repository markers were found')
    except AutopypathError:
        pass  # Expected because no repository markers exist


def test_restore_sys_path(tmp_path: Path) -> None:
    """Test that sys.path is restored on call to _config.restore_sys_path()."""
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    context_file = root_path / 'some_file.txt'
    context_file.write_text('Just a test file.')
    git_path = root_path / '.git'
    git_path.mkdir()
    src_path = root_path / 'src'
    src_path.mkdir()

    # dry run - sys.path should not be modified
    try:
        sys.path = _ORIGINAL_SYS_PATH.copy()
        config = _ConfigPyPath(
            context_file=context_file,
            load_strategy='prepend',
            path_resolution_order=['manual'],
            paths=['src'],
            repo_markers={'.git': 'dir'},
            dry_run=True,
        )

        assert sys.path == _ORIGINAL_SYS_PATH, 'RESTORE_SYS_PATH_001 sys.path should be unchanged in dry run mode'
        config.restore_sys_path()
        assert sys.path == _ORIGINAL_SYS_PATH, (
            'RESTORE_SYS_PATH_002 sys.path should be still be unchanged after restore_sys_path() is called'
        )

    finally:
        sys.path = _ORIGINAL_SYS_PATH.copy()

    # live - sys.path should be modified and then restored
    try:
        sys.path = _ORIGINAL_SYS_PATH.copy()
        config = _ConfigPyPath(
            context_file=context_file,
            load_strategy='prepend',
            path_resolution_order=['manual'],
            paths=['src'],
            repo_markers={'.git': 'dir'},
        )

        assert len(sys.path) == len(_ORIGINAL_SYS_PATH) + 1, (
            'RESTORE_SYS_PATH_001 sys.path should have one additional entry after prepend strategy is applied'
        )
        config.restore_sys_path()
        assert sys.path == _ORIGINAL_SYS_PATH, (
            'RESTORE_SYS_PATH_002 sys.path should be restored to its original state after restore_sys_path() is called'
        )

    finally:
        sys.path = _ORIGINAL_SYS_PATH.copy()


def test_updated_sys_path_property(tmp_path: Path) -> None:
    """Test that _ConfigPyPath.updated_sys_path property is unchanged in dry run mode.
    It should return the expected sys.path
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    context_file = root_path / 'some_file.txt'
    context_file.write_text('Just a test file.')
    git_path = root_path / '.git'
    git_path.mkdir()
    src_path = root_path / 'src'
    src_path.mkdir()

    try:
        sys.path = _ORIGINAL_SYS_PATH.copy()
        config = _ConfigPyPath(
            context_file=context_file,
            load_strategy='prepend',
            path_resolution_order=['manual'],
            paths=['src'],
            repo_markers={'.git': 'dir'},
            dry_run=True,
        )

        updated_sys_path = config.updated_sys_path
        assert len(updated_sys_path) == len(_ORIGINAL_SYS_PATH), (
            'UPDATED_SYS_PATH_001 updated_sys_path should have one additional entry after prepend strategy is applied'
        )

    finally:
        sys.path = _ORIGINAL_SYS_PATH.copy()


def test_path_order_resolution(tmp_path: Path) -> None:
    """Tests that _ConfigPyPath correctly resolves configuration
    for path resolution order specified in various configuration sources.

    Uses a temporary directory to simulate a repository with an autopypath.toml file.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    git_path = root_path / '.git'
    git_path.mkdir()

    manual_src_path = root_path / 'manual_src'
    manual_src_path.mkdir()

    pyproject_src_path = root_path / 'pyproject_src'
    pyproject_src_path.mkdir()

    autopypath_src_path = root_path / 'autopypath_src'
    autopypath_src_path.mkdir()

    # Create pyproject.toml with some configuration
    pyproject_path = root_path / 'pyproject.toml'
    pyproject_path.write_text("""
[tool.autopypath]
load_strategy = "prepend"
path_resolution_order = ["pyproject"]
repo_markers = {".git" = "dir"}
paths=["pyproject_src"]
""")

    # Create autopypath.toml with some configuration
    autopypath_path = root_path / 'autopypath.toml'
    autopypath_path.write_text("""
[tool.autopypath]
load_strategy = "prepend"
path_resolution_order = ["autopypath", "pyproject"]
repo_markers = {".git" = "dir"}
paths=["autopypath_src"]
""")

    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        load_strategy='prepend',
        path_resolution_order=['manual', 'autopypath', 'pyproject'],
        paths=['manual_src'],
        repo_markers={'.git': 'dir'},
    )

    # manual config should be highest precedence and thus first in sys.path
    assert sys.path[0] == str(manual_src_path), (
        f'PATH_ORDER_RESOLUTION_001 manual_src should be first in updated_sys_path: {sys.path!r}'
    )

    # autopypath.toml config should be second precedence
    assert sys.path[1] == str(autopypath_src_path), (
        f'PATH_ORDER_RESOLUTION_002 autopypath_src should be second in updated_sys_path: {sys.path!r}'
    )

    # pyproject.toml config should be third precedence
    assert sys.path[2] == str(pyproject_src_path), (
        f'PATH_ORDER_RESOLUTION_003 pyproject_src should be third in updated_sys_path: {sys.path!r}'
    )

    config.restore_sys_path()

    # reverse the order to verify it wasn't a fluke
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        load_strategy='prepend',
        path_resolution_order=['pyproject', 'autopypath', 'manual'],
        paths=['manual_src'],
        repo_markers={'.git': 'dir'},
    )

    # pyproject.toml config should be first precedence
    assert sys.path[0] == str(pyproject_src_path), (
        f'PATH_ORDER_RESOLUTION_006 pyproject_src should be first in updated_sys_path: {sys.path!r}'
    )
    # autopypath.toml config should be third precedence
    assert sys.path[1] == str(autopypath_src_path), (
        f'PATH_ORDER_RESOLUTION_007 autopypath_src should be second in updated_sys_path: {sys.path!r}'
    )
    # manual config should be fourth precedence
    assert sys.path[2] == str(manual_src_path), (
        f'PATH_ORDER_RESOLUTION_008 manual_src should be third in updated_sys_path: {sys.path!r}'
    )

    sys.path = []  # Clear sys.path for next test

    # Finally, test with only autopypath.toml and pyproject.toml configs
    # manual is not in the resolution order
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        load_strategy='prepend',
        path_resolution_order=['autopypath', 'pyproject'],
        repo_markers={'.git': 'dir'},
    )
    # autopypath.toml config should be first precedence
    assert sys.path[0] == str(autopypath_src_path), (
        f'PATH_ORDER_RESOLUTION_009 autopypath_src should be first in updated_sys_path: {sys.path!r}'
    )
    # pyproject.toml config should be second precedence
    assert sys.path[1] == str(pyproject_src_path), (
        f'PATH_ORDER_RESOLUTION_010 pyproject_src should be second in updated_sys_path: {sys.path!r}'
    )

    assert len(sys.path) == 2, f'PATH_ORDER_RESOLUTION_011 updated_sys_path should have length 2: {sys.path!r}'

    sys.path = _ORIGINAL_SYS_PATH.copy()


def test_load_strategy_precedence(tmp_path: Path) -> None:
    """Tests that _ConfigPyPath respects the load_strategy precedence
    in the hierarchical configuration resolution.

    manual > autopypath.toml > pyproject.toml  > defaults.

    Uses a temporary directory to simulate a repository with an autopypath.toml file.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    git_path = root_path / '.git'
    git_path.mkdir()
    src_path = root_path / 'src'
    src_path.mkdir()

    # Create pyproject.toml with load_strategy configuration
    pyproject_path = root_path / 'pyproject.toml'
    pyproject_path.write_text("""
[tool.autopypath]
load_strategy = "replace"
path_resolution_order = ["pyproject"]
repo_markers = {".git" = "dir"}
paths=["src"]
""")
    # Create autopypath.toml with load_strategy configuration
    autopypath_path = root_path / 'autopypath.toml'
    autopypath_path.write_text("""
[tool.autopypath]
load_strategy = "prepend"
path_resolution_order = ["autopypath", "pyproject"]
repo_markers = {".git" = "dir"}
paths=["src"]
""")
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        load_strategy='prepend_highest_priority',
        path_resolution_order=['manual', 'autopypath', 'pyproject'],
        dry_run=True,
        paths=['src'],
        repo_markers={'.git': 'dir'},
    )

    # manual config should be highest precedence and hence the winner
    assert config.load_strategy == 'prepend_highest_priority', (
        'LOAD_STRATEGY_PRECEDENCE_001 load_strategy should be prepend_highest_priority from manual config'
    )

    # autopypath.toml config should winner as second precedence without a manual load_strategy
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        path_resolution_order=['autopypath', 'pyproject'],
        dry_run=True,
        paths=['src'],
        repo_markers={'.git': 'dir'},
    )
    assert config.load_strategy == 'prepend', (
        'LOAD_STRATEGY_PRECEDENCE_002 load_strategy should be prepend from autopypath.toml config'
    )

    # pyproject.toml config should be winner as third precedence without manual or autopypath load_strategy
    autopypath_path.unlink()  # Remove autopypath.toml to test pyproject.toml precedence

    # Create autopypath.toml without load_strategy configuration
    autopypath_path = root_path / 'autopypath.toml'
    autopypath_path.write_text("""
[tool.autopypath]
path_resolution_order = ["autopypath", "pyproject"]
repo_markers = {".git" = "dir"}
paths=["src"]
""")
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        path_resolution_order=['autopypath', 'pyproject'],
        dry_run=True,
        paths=['src'],
        repo_markers={'.git': 'dir'},
    )
    assert config.load_strategy == 'replace', (
        'LOAD_STRATEGY_PRECEDENCE_003 load_strategy should be replace from pyproject.toml config'
    )

    # finally, defaults should be winner without any load_strategy configured

    autopypath_path.unlink()  # Remove autopypath.toml to test default precedence
    pyproject_path.unlink()  # Remove pyproject.toml to test default precedence

    # default load_strategy should be winner without manual, autopypath, or pyproject load_strategy
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        path_resolution_order=['autopypath', 'pyproject'],
        dry_run=True,
        paths=['src'],
        repo_markers={'.git': 'dir'},
    )
    assert config.load_strategy == defaults._LOAD_STRATEGY, (
        'LOAD_STRATEGY_PRECEDENCE_004 load_strategy should be prepend from default configuration'
    )


def test_path_resolution_order_precedence(tmp_path: Path) -> None:
    """Tests that _ConfigPyPath respects the path_resolution_order precedence
    in the hierarchical configuration resolution.

    manual > autopypath.toml > pyproject.toml  > defaults.

    Uses a temporary directory to simulate a repository with an autopypath.toml file.
    """
    root_path = tmp_path / 'repo'
    root_path.mkdir()
    root_path.joinpath('some_file.txt').write_text('Just a test file.')
    git_path = root_path / '.git'
    git_path.mkdir()
    src_path = root_path / 'src'
    src_path.mkdir()

    # Create pyproject.toml with path_resolution_order configuration
    pyproject_path = root_path / 'pyproject.toml'
    pyproject_path.write_text("""
[tool.autopypath]
path_resolution_order = ["pyproject"]
repo_markers = {".git" = "dir"}
paths=["src"]
""")
    # Create autopypath.toml with path_resolution_order configuration
    autopypath_path = root_path / 'autopypath.toml'
    autopypath_path.write_text("""
[tool.autopypath]
path_resolution_order = ["autopypath", "pyproject"]
repo_markers = {".git" = "dir"}
paths=["src"]
""")
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        path_resolution_order=['manual', 'autopypath', 'pyproject'],
        dry_run=True,
        paths=['src'],
        repo_markers={'.git': 'dir'},
    )
    # manual config should be highest precedence and hence the winner
    assert config.path_resolution_order == ('manual', 'autopypath', 'pyproject'), (
        'PATH_RESOLUTION_ORDER_PRECEDENCE_001 path_resolution_order should be '
        f'["manual", "autopypath", "pyproject"] from manual config: {config.path_resolution_order}'
    )
    # autopypath.toml config should winner as second precedence without a manual path_resolution_order
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        path_resolution_order=['autopypath', 'pyproject'],
        dry_run=True,
        paths=['src'],
        repo_markers={'.git': 'dir'},
    )
    assert config.path_resolution_order == ('autopypath', 'pyproject'), (
        'PATH_RESOLUTION_ORDER_PRECEDENCE_002 path_resolution_order should be '
        '["autopypath", "pyproject"] from autopypath.toml config'
    )
    # pyproject.toml config should be winner as third precedence without manual or autopypath path_resolution_order
    autopypath_path.unlink()  # Remove autopypath.toml to test pyproject.toml precedence
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        path_resolution_order=['pyproject'],
        dry_run=True,
        paths=['src'],
        repo_markers={'.git': 'dir'},
    )
    assert config.path_resolution_order == ('pyproject',), (
        'PATH_RESOLUTION_ORDER_PRECEDENCE_003 path_resolution_order should be ["pyproject"] from pyproject.toml config'
    )
    # finally, defaults should be winner without any path_resolution_order configured
    pyproject_path.unlink()  # Remove pyproject.toml to test default precedence
    config = _ConfigPyPath(
        context_file=root_path / 'some_file.txt',
        dry_run=True,
        paths=['src'],
        repo_markers={'.git': 'dir'},
    )
    assert config.path_resolution_order == defaults._PATH_RESOLUTION_ORDER, (
        'PATH_RESOLUTION_ORDER_PRECEDENCE_004 path_resolution_order should be '
        f'{defaults._PATH_RESOLUTION_ORDER} from default configuration'
    )



"""Tests for autopypath._config_py_path._config._manual module.

The Manual config class is just no changes subclass of
the base Config class, so this test just verifies that the
class can be instantiated as ManualConfig and is a subclass
of Config.

All the real logic is tested in tests/_config_py_path/test_config.py
"""

from autopypath import _defaults
from autopypath._config_py_path._config import _Config, _DefaultConfig


def test_default_config_init() -> None:
    """Test that DefaultConfig can be instantiated with default parameters."""
    config = _DefaultConfig()

    assert isinstance(config, _DefaultConfig), 'DEFAULT_001 Failed to instantiate DefaultConfig'
    assert issubclass(_DefaultConfig, _Config), 'DEFAULT_002 DefaultConfig is not a subclass of Config'
    assert config.load_strategy == _defaults._LOAD_STRATEGY, 'DEFAULT_003 Load strategy does not match default'
    assert config.paths == _defaults._PATHS, 'DEFAULT_004 Paths do not match default'
    assert config.repo_markers == _defaults._REPO_MARKERS, 'DEFAULT_005 Repo markers do not match default'
    assert config.path_resolution_order == _defaults._PATH_RESOLUTION_ORDER, (
        'DEFAULT_006 Path resolution order does not match default'
    )


def test_repr() -> None:
    """Test Config __repr__ method."""
    # Note - use of eval on the repr output to reconstitute the Config instance
    # and compare to the original. This is safe in this controlled test context
    # where we know the input values and expected output. In general, eval
    # should be avoided due to security risks. IOW: This is a test-only use of eval.
    config = _DefaultConfig()

    reconstituted_config = eval(repr(config))
    assert reconstituted_config == config, (
        'REPR_001 Reconstituted Config from __repr__ does not match original. '
        f'Original: {config!r}, Reconstituted: {reconstituted_config!r}'
    )



"""Default configuration values for autopypath package."""

from pathlib import Path
from types import MappingProxyType
from typing import Final, Union

from ._load_strategy import _LoadStrategy
from ._marker_type import _MarkerType
from ._path_resolution import _PathResolution
from ._types import LoadStrategyLiterals, PathResolutionLiterals, RepoMarkerLiterals

__all__ = []

_REPO_MARKERS: Final[MappingProxyType[str, Union[_MarkerType, RepoMarkerLiterals]]] = MappingProxyType(
    {
        'pyproject.toml': _MarkerType.FILE,
        'autopypath.toml': _MarkerType.FILE,
        '.git': _MarkerType.DIR,  # Git repository marker
        '.hg': _MarkerType.DIR,  # Mercurial repository marker
        '.svn': _MarkerType.DIR,  # Subversion repository marker
        '.bzr': _MarkerType.DIR,  # Bazaar repository marker
        '.cvs': _MarkerType.DIR,  # CVS repository marker
        '_darcs': _MarkerType.DIR,  # Darcs repository marker
        '.fossil': _MarkerType.DIR,  # Fossil repository marker
    }
)
"""Default repository markers used to identify the repository root. The presence of
any of these files or directories indicates the root of the project repository.

Default markers are:
- ``pyproject.toml``: Indicates the repository root by the presence of this file.
- ``autopypath.toml``: Indicates the repository root by the presence of this file.

    Note that there is special behavior for this marker if found. If it changes
    the repo_markers settings, the directory it is found in is evaluated using the new settings
    and that may result in a different repository root being identified.

    This allows autopypath.toml to be used to directly identify the repository root
    OR to customize the repo markers used to actually identify the repository root.

    Only the first autopypath.toml file found when searching upwards from
    the starting directory is used to allow hierarchical configurations.
    Additional autopypath.toml files found higher up the directory tree are ignored.

- ``.git``: Indicates a Git repository root by the presence of this directory.
- ``.hg``: Indicates a Mercurial repository root by the presence of this directory.
- ``.svn``: Indicates a Subversion repository root by the presence of this directory.
- ``.bzr``: Indicates a Bazaar repository root by the presence of this directory.
- ``.cvs``: Indicates a CVS repository root by the presence of this directory.
- ``_darcs``: Indicates a Darcs repository root by the presence of this directory.
- ``.fossil``: Indicates a Fossil repository root by the presence of this directory.

These can be overridden by the `repo_markers` parameter to :func:`configure_pypath`.
They can also be configured in the `pyproject.toml` file under the
`[tool.autopypath]` section.

The order for checking these markers is as follows. Resolution stops
at the first level the markers are defined.

If manual repo markers are provided via :func:`configure_pypath`, those are the only markers checked.

If no manual repo markers are provided but `pyproject.toml` defines `repo_markers`, those are used instead.

If neither manual repo markers nor `pyproject.toml` defines `repo_markers`, the default markers listed above are used

The order is:

custom markers
--------------
.. code-block:: python
    from autopypath.custom import configure_pypath
    configure_pypath(
        repo_markers={'custom_marker.txt': 'file', 'special_dir': 'dir'}
    )

pyproject.toml file or .git directory
-------------------------------------

.. code-block:: toml
    [tool.autopypath]
    repo_markers = { 'pyproject.toml' = 'file', '.git' = 'dir' }

default markers
-----------------

The default markers are:

.. code-block:: python

    {
        'pyproject.toml': 'file',
        'autopypath.toml': 'file',
        '.git': 'dir',
        '.hg': 'dir',
        '.svn': 'dir',
        '.bzr': 'dir',
        '.cvs': 'dir',
        '_darcs': 'dir',
        '.fossil': 'dir',
    }

"""

_PATH_RESOLUTION_ORDER: Final[tuple[Union[_PathResolution, PathResolutionLiterals], ...]] = (
    _PathResolution.MANUAL,
    _PathResolution.AUTOPYPATH,
    _PathResolution.PYPROJECT,
)

"""Default resolution order for :data:`sys.path` sources.

This is used if there is no specific resolution order provided in pyproject.toml
or other configuration.

It be overridden by the `path_resolution_order` parameter to `configure_pypath()`
or by configuring it in `pyproject.toml` under the
`[tool.autopypath]` section.

If overridden, it should be provided as a sequence of strings.
The overrides will replace the entire default resolution order
and define a new prioritization order for resolving :data:`sys.path` sources.


Overrides can use any combination of the following values.
- `manual`: Paths provided directly via the `paths` parameter to `configure_pypath()`.
- `autopypath`: Paths specified in an `autopypath.toml` file.
- `pyproject`: Paths specified in the `pyproject.toml` file in the repository root.

Override Example
----------------

Manual Override
~~~~~~~~~~~~~~~

.. code-block:: python
    from autopypath.custom import configure_pypath, PathResolution
    configure_pypath(
        path_resolution_order=['manual', 'autopypath', 'pyproject']
    )

pyproject.toml or autopypath.toml Override
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

You can specify the resolution order in `pyproject.toml` or `autopypath.toml`
in a `[tool.autopypath]` section

Omitted path resolution types will not be used for resolution at runtime.

.. code-block:: toml
    [tool.autopypath]
    path_resolution_order = ['manual', 'autopypath', 'pyproject']

These examples apply the following path prioritization order:

1. `paths` manually specified in a :func:`~autopypath.custom.configure_pypath` call (highest priority)
2. Paths specified in an `autopypath.toml` file
3. Paths specified in a `pyproject.toml` file in the repository root
"""

_LOAD_STRATEGY: Union[_LoadStrategy, LoadStrategyLiterals] = _LoadStrategy.PREPEND
"""Default load strategy for :data:`sys.path` sources.

This is used if there is no specific load strategy provided in pyproject.toml
or other configuration.

The available load strategies are defined as follows:

- **prepend** (default)

    Combines paths from all sources.

    This allows a developer's local shell environment to supplement the
    paths defined in the project's configuration files. Paths are added
    to the front of `sys.path` in the order of priority.

- **prepend_highest_only**

    Uses paths from the most specific source found (highest priority) only,
    ignoring others. This ensures a predictable path environment.

- **replace**
    Replaces the entire `sys.path` with the merged paths from all sources.
    This may break standard library and installed package imports,
    so it should be used with caution. Don't use this unless you really know
    what you're doing: you will likely need to add back standard library
    and installed package paths: zeroing out `sys.path` is generally not recommended.

    If you have to ask, you probably shouldn't be using this option. It is
    intended for very specialized use cases only.


These can be overridden by the `load_strategy` parameter to :func:`configure_pypath`.
They can also be configured in the `pyproject.toml` file under the
`[tool.autopypath]` section.

The order for defining these load strategies is as follows. Definition stops
at the first level the strategies are defined at.

If a manual load strategies is provided via :func:`configure_pypath`, that is the stragegy used.

If no manual load strategy is provided but `pyproject.toml` defines `load_strategy`, that is used instead.

If `load_strategy` is not defined either manually or via `pyproject.toml`, the default strategy listed above is used

The order is:

prepend
-------
.. code-block:: python
    from autopypath.custom import configure_pypath, MarkerType
    configure_pypath(
        load_strategy='prepend'
    )

prepend highest only
--------------------

.. code-block:: toml
    [tool.autopypath]
    load_strategy = 'prepend_highest_priority'

default markers
-----------------

The default strategy as defined above.

"""

_PATHS: Final[tuple[Path, ...]] = (Path('src'), Path('tests'), Path('lib'), Path('src/test'))
"""Default paths to add to sys.path relative to the repository root.

The default paths, in order of priority, are:
- `src`: Common source directory for project code.
- `tests`: Common source directory for test code.
- `lib`: Common source directory for library code.
- `src/test`: Alternate test source directory.

These directories are added to :data:`sys.path` if they exist in the repository root.

They can be customized by the `paths` parameter to `configure_pypath()` or
configured in a `pyproject.toml` or `autopypath.toml` file in the `[tool.autopypath]` section.

This is just the 'out-of-the-box' default configuration and is intended to cover
the most common project layouts. You can (and **should**) customize it as needed for your project.

If a configured path does not exist or is not a directory, it will be logged at 'INFO' level and skipped.

.. code-block:: toml
    [tool.autopypath]
    paths = ['src', 'tests']

"""
